from save_to_db.core import signals

from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestSignals(TestBase):
    
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    def test_db_persist_signals(self):
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
        
        persister = self.persister
        
        signal_calls = {
            'before_one': 0,
            'before_two': 0,
            'after_one': 0,
            'after_two': 0,
        }
        
        item_one = ItemGeneralOne(f_integer=1)
        item_one['two_x_x'](f_integer=1)
        persister.persist(item_one)
        self.assertEqual(signal_calls, {'before_one': 0,
                                        'before_two': 0,
                                        'after_one': 0,
                                        'after_two': 0,
                                        })
        
        @signals.before_db_persist.register
        def before_db_persist_signal_one(item, item_structure):
            nonlocal expected_item, expected_item_structure
            self.assertIs(item, expected_item)
            self.assertEqual(item_structure, expected_item_structure)
            self.assertIsInstance(item_structure, dict)
            signal_calls['before_one'] += 1
        
        expected_item = item_one
        expected_item_structure = item_one.process()
        persister.persist(item_one)
        self.assertEqual(signal_calls, {'before_one': 1,
                                        'before_two': 0,
                                        'after_one': 0,
                                        'after_two': 0,
                                        })
        
        def before_db_persist_signal_two(item, item_structure):
            nonlocal expected_item, expected_item_structure
            self.assertIs(item, expected_item)
            self.assertEqual(item_structure, expected_item_structure)
            self.assertIsInstance(item_structure, dict)
            signal_calls['before_two'] += 1
        signals.before_db_persist.register(before_db_persist_signal_two)
        
        persister.persist(item_one)
        self.assertEqual(signal_calls, {'before_one': 2,
                                        'before_two': 1,
                                        'after_one': 0,
                                        'after_two': 0,
                                        })
        
        @signals.after_db_persist.register
        def after_db_persist_signal_one(top_item, items, models):
            nonlocal expected_item, signal_items, signal_models
            self.assertIs(top_item, expected_item)
            signal_items, signal_models = items, models
            signal_calls['after_one'] += 1
        
        signal_items, signal_models = None, None
        returned_items, returned_models = persister.persist(item_one)
        self.assertIs(returned_items, signal_items)
        self.assertIs(returned_models, signal_models)
        self.assertEqual(signal_calls, {'before_one': 3,
                                        'before_two': 2,
                                        'after_one': 1,
                                        'after_two': 0,
                                        })
        
        def after_db_persist_signal_two(top_item, items, models):
            nonlocal expected_item, signal_items, signal_models
            self.assertIs(top_item, expected_item)
            signal_items, signal_models = items, models
            signal_calls['after_two'] += 1
        signals.after_db_persist.register(after_db_persist_signal_two)
            
        returned_items, returned_models = persister.persist(item_one)
        self.assertIs(returned_items, signal_items)
        self.assertIs(returned_models, signal_models)
        self.assertEqual(signal_calls, {'before_one': 4,
                                        'before_two': 3,
                                        'after_one': 2,
                                        'after_two': 1,
                                        })
        
        # unregister
        signals.before_db_persist.unregister(before_db_persist_signal_one)
        signals.after_db_persist.unregister(after_db_persist_signal_one)
        
        returned_items, returned_models = persister.persist(item_one)
        self.assertIs(returned_items, signal_items)
        self.assertIs(returned_models, signal_models)
        self.assertEqual(signal_calls, {'before_one': 4,
                                        'before_two': 4,
                                        'after_one': 2,
                                        'after_two': 2,
                                        })
            
    def test_item_dropped_signal(self):
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
        
        persister = self.persister
        
        
        item_one = ItemGeneralOne(f_integer='10')
        # dropped because cannot be created
        item_two_dropped = item_one['two_x_x'].gen(f_string='str-20')
        # update only mode
        item_one_dropped = item_two_dropped['one_x_1'](f_integer='30')
        item_one_dropped.update_only_mode = True
        # get only mode
        item_one_dropped_2 = item_two_dropped['one_x_x'].gen(f_integer='40')
        item_one_dropped_2.get_only_mode = True
        
        items_to_reasons = [
            [
                item_two_dropped,
                signals.item_dropped.reason_cannot_create_not_enough_data,
            ],
            [
                item_one_dropped,
                signals.item_dropped.reason_cannot_create_update_only_mode,
            ],
            [
                item_one_dropped_2,
                signals.item_dropped.reason_cannot_create_get_only_mode,
            ],
        ]
        
        @signals.item_dropped.register
        def item_dropped_listener(item, reason):
            nonlocal items_to_reasons
            
            for entry in items_to_reasons:
                if entry[0] is item:
                    try:
                        self.assertEqual(entry[1], reason, item)
                    except:
                        print(reason, entry[0].update_only_mode, entry[0]['f_integer'])
                        raise
                    items_to_reasons.remove(entry)
                    break
            else:
                raise Exception('Entry not found: [{}, {}]'.format(
                    item, reason))
            
        
        items, model_list = persister.persist(item_one)
        self.assertEqual(len(items), 1)
        self.assertEqual(len(model_list), 1)
        self.assertEqual(len(model_list[0]), 1)
        self.assertIs(items[0], item_one)
        
        self.assertEqual(len(items_to_reasons), 0)
        
        item_ones = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(item_ones), 1)
        item_twos = self.get_all_models(self.ModelGeneralTwo)
        self.assertEqual(len(item_twos), 0)
        
        
            
            
            
            
            
        


import inspect
from save_to_db import tests
from unittest import TestCase


def load(storage):
    # loading all tests into current name space
    for test_class_name, test_class in inspect.getmembers(tests):
        if not inspect.isclass(test_class) or not issubclass(test_class, TestCase):
            continue
        
        storage[test_class_name] = test_class
        




""" This module contains exceptions for :py:class:`~save_to_db.core.scope.Scope`
class.
"""

class ScopeException(Exception):
    """ General exception for :py:class:`~save_to_db.core.scope.Scope` class.
    """


class ItemClsAlreadyScoped(ScopeException):
    """ Raised when an already scoped item is scoped again. """


class ScopeIdCannotBeNone(ScopeException):
    """ Raised when trying to set Scope `collection_id` to `None`. """


from save_to_db.exceptions import (CannotMergeModels,
                                   CannotMergeModelsForItem,
                                   MultipleModelsMatch,)
from save_to_db.core.item_base import ItemBase
from save_to_db.adapters.utils.relation_type import RelationType

from .process_keeper import ProcessKeeper



class ModelUpdater(object):
    
    def __init__(self, item_structure, db_adapter):
        self.process_keeper = ProcessKeeper(item_structure)
        self.db_adapter = db_adapter
        self.created_models = []
        # for x-to-many relationships we have to keep track of missing values
        # as during model updated after first model is set, relationship is
        # no longer empty, the rest of the child models won't set 
        self.no_value_models = []
        self.no_value_fields = []
    
    def add_created_model(self, model):
        self.created_models.append(model)
    
    def is_model_created(self, model):
        return model in self.created_models
    
    #--- checkers --------------------------------------------------------------
    
    def can_set_model_field(self, item, model, field_name, value):
        """ Checks if we can set given `value` for the given `model` using
        given `field_name`.
        
        .. note::
            `model` can be `None` if the corresponding model for the item was
            not yet created.
            
        :param item: Item used to populated `model`.
        :param model: ORM model that corresponds to the `item`.
        :param field_name: Field name used to set value for the `model`.
        :param value: Value to set. For x-to-many relationships only a single
            value must be used.
        :returns: `True` if the `value` can be set, `False` otherwise.
        """
        # checking for get_only_mode
        if item.get_only_mode or \
                (isinstance(value, ItemBase) and value.get_only_mode):
            return False
        
        # checking if the field can be always set
        if field_name not in item.norewrite_fields or \
                item.norewrite_fields[field_name] is None:
            return True
        
        created = model is None or self.is_model_created(model)
        rewrite_none = item.norewrite_fields[field_name]
        
        # not a relation
        if field_name not in item.relations:
            if created:
                return True
            model_value = getattr(model, field_name)
            return False if model_value is not None else rewrite_none
        
        # value is a related model
        elif created and (value is None or self.is_model_created(value)):
            return True
        
        relation_type = item.relations[field_name]['relation_type']
        
        # for x-to-many relationships we can add to many a newly created model
        if value:
            if relation_type.is_many_to_x() and created:
                return True
            if relation_type.is_x_to_many() and \
                    (value is None or self.is_model_created(value)):
                return True
        
        # at least one of the model already existed, we can set only if
        # we can rewrite nones and the values are `None`
        if rewrite_none is False:
            return False
            
        # checking values
        def model_had_value(relation_type, model, field_name):
            if model in self.no_value_models:
                index = self.no_value_models.index(model)
                if field_name in self.no_value_fields[index]:
                    return False
            
            result = __model_had_value(relation_type, model, field_name)
            
            if not result:
                if model not in self.no_value_models:
                    self.no_value_models.append(model)
                    self.no_value_fields.append([field_name])
                else:
                    index = self.no_value_models.index(model)
                    self.no_value_fields[index].append(field_name)
            return result
        
            
        def __model_had_value(relation_type, model, field_name):    
            if not hasattr(model, field_name):
                return False
            if relation_type.is_x_to_one():
                model_value = getattr(model, field_name)
                return model_value is not None
            else:
                return self.db_adapter.related_x_to_many_exists(
                    model, field_name)
        
        # this side
        if model and model_had_value(relation_type, model, field_name):
            return False
        
        # that side
        if value is not None:
            reverse_key = item.relations[field_name]['reverse_key']
            if reverse_key and model_had_value(relation_type.reverse(),
                                               value, reverse_key):
                return False
        
        return True
    
    
    def can_create_model(self, item, fkeys):
        """ Checks that a new row can be created in database for
        the item.
        
        .. note::
            For relations fields we check that it is in `fkeys`. For normal
            fields we just check that it is present and not `None`.
        
        :param item: Item for which a corresponding model should be created.
        :param fkeys: A `dict` where keys are relation field names and values
            are ORM models that were got from or created in database.
        :returns: `True` if a corresponding row for the item can be created in
            database, `False` otherwise.
        """
        if item.creators is None or item.get_only_mode or item.update_only_mode:
            return False
        if not item.creators:
            return True
        
        for group in item.creators:
            can_create = True
            for field_name in group:
                if field_name in item.relations:
                    # checking can set on the other side of relationship
                    if field_name not in fkeys:
                        can_create = False
                        break
                    for f_model in fkeys[field_name]:
                        if not self.can_set_model_field(item, None, field_name,
                                                        f_model):
                            can_create = False
                            break
                    if not can_create:
                        break
                        
                # always can set normal fields when creating
                elif field_name not in item.data:
                    can_create = False
                    break
                
            if can_create:
                return True
        return False
    
    #--- modifiers -------------------------------------------------------------
    
    def clear_related_models(self, item, model, fkey):
        """ Removes all referenced models by the foreign key `fkey` of `model`
        if possible.
        
        :param item: Item used to get or create the `model`.
        :param model: model created or loaded from database using the `item`.
        :param fkey: x-to-many foreign key that needs to get empty. 
        """
        relation = item.relations[fkey]
        
        if not self.db_adapter.REVERSE_MODEL_AUTOUPDATE_SUPPORTED and \
                relation['reverse_key'] and relation['item_cls'] and \
                relation['relation_type'] is RelationType.ONE_TO_MANY:
            item_cls = relation['item_cls']
            
            models_to_check = []
            for item_track in self.process_keeper[item_cls]:
                models_to_check.extend(item_track.models)
            contained_models = self.db_adapter.related_x_to_many_contains(
                model, fkey, models_to_check)
             
            for contained_model in contained_models:
                self.set_related_model(item_track.item,
                                       contained_model,
                                       relation['reverse_key'],
                                       None)
    
        self.db_adapter.clear_related_models(model, fkey)
        
        # clearing model_unrefs
        model_unrefs = item.metadata['model_unrefs']
        if fkey in model_unrefs:
            model_unrefs[fkey].reset()
        
    
    def set_related_model(self, item, model, fkey, fmodel):
        """ Sets a model `fmodel` as a related model of `model` under
        foreign key `fkey` if possible.
        
        :param item: Item used to get or create the `model`.
        :param model: model created or loaded from database using the `item`.
        :param fkey: foreign key used to reference `fmodel`.
        :param fmodel: referenced foreign model. 
        """
        if not self.can_set_model_field(item, model, fkey, fmodel):
            return
            
        def clean_1_to_1_relation(model, fkey, fmodel, item_cls, reverse_key):
            if reverse_key and hasattr(model, fkey):
                related_model = getattr(model, fkey)
                if related_model and related_model != fmodel:
                    setattr(related_model, reverse_key, None)
                    self.db_adapter.save_model(related_model)
             
            for item_track in self.process_keeper[item_cls]:
                found = False
                for track_model in item_track.models:
                    if track_model == model:
                        continue
                    if hasattr(track_model, fkey) and \
                            getattr(track_model, fkey) == fmodel:
                        found = True
                        setattr(track_model, fkey, None)
                        self.db_adapter.save_model(track_model)
                        break
                 
                if found:
                    break
                 
        relation = item.relations[fkey]
        if relation['relation_type'] is RelationType.ONE_TO_ONE and \
                ((not hasattr(model, fkey) and fmodel is not None) or
                 (hasattr(model, fkey) and getattr(model, fkey) != fmodel)):
            # forward relation
            clean_1_to_1_relation(model, fkey, fmodel,
                                  type(item), relation['reverse_key'])
            # backward
            if relation['reverse_key']:
                clean_1_to_1_relation(fmodel, relation['reverse_key'], model,
                                      relation['item_cls'], fkey)
        
        setattr(model, fkey, fmodel)
    
    
    def add_related_models(self, item, model, fkey, fmodels):
        """ Adds related models `fmodels` to x-to-many relationship of `model`
        using foreign key `fkey`.
        
        :param item: Item used to get or create the `model`.
        :param model: model created or loaded from database using the `item`.
        :param fkey: foreign key used to reference `fmodels`.
        :param fmodels: referenced models.
        """
        # no need to updated other models:
        #    1. If it is many-to-many relationship, then calling this function
        #       does not mean that other relationships must be nullified
        #    2. If it one-to-many relationship, then foreign key is on the
        #       `fmodels` models.
        filtered_models = [
            fmodel for fmodel in fmodels
            if self.can_set_model_field(item, model, fkey, fmodel)
        ]
        self.db_adapter.add_related_models(model, fkey, filtered_models)
        
        # saving to model_unrefs
        model_unrefs = item.metadata['model_unrefs']
        if fkey in model_unrefs:
            for fmodel in fmodels:
                if self.is_model_created(fmodel):
                    self.db_adapter.save_model(fmodel)
                model_unrefs[fkey].collect_model(fmodel)
    
    #---------------------------------------------------------------------------
    
    def merge_models_for_item(self, item, models):
        if not models:
            return None
        
        # sorting models to always get the same result after merging
        models = list(models)
        pk_list = list(self.db_adapter.get_primary_key_names(type(models[0])))
        pk_list.sort()
        get_model_key = lambda model: tuple(getattr(model, field)
                                            for field in pk_list)
        models.sort(key=get_model_key)
        
        # `before_models_merge` hook
        item.before_models_merge(models)
        
        try:
            result_model = self.db_adapter.merge_models(models)
        except CannotMergeModels as e:
            raise CannotMergeModelsForItem(item, *e.args)
        
        # `after_models_merge` hook
        item.after_models_merge(result_model)
        
        return result_model
    
    
    def update_model_fields(self, item_track, model):
        created = self.is_model_created(model)
        
        for field_name in item_track.item.data:
            if field_name in item_track.item.fields:
                if self.can_set_model_field(item_track.item, model, field_name,
                                            item_track.item[field_name]):
                    setattr(model, field_name, item_track.item[field_name])
            
        # many to one before saving
        for fkey in item_track.fkeys:
            relation_type = item_track.item.relations[fkey]['relation_type']
            if relation_type.is_x_to_one():
                self.set_related_model(item_track.item, model, fkey,
                                       item_track.fkeys[fkey][0])
        
        # for the rest model has to be saved
        if created and self.db_adapter.SAVE_MODEL_BEFORE_COMMIT:
            self.db_adapter.save_model(model)
    
        for fkey, fmodels in item_track.fkeys.items():
            relation_type = item_track.item.relations[fkey]['relation_type']
            if not relation_type.is_x_to_one():
                self.add_related_models(item_track.item,  model, fkey, fmodels)
    
    
    def update_relationships(self, items, models):
        """ Completes relationships on the `items`, e.g. sets reverse
        relationships where possible.
        
        :param items: All the items that we deal with. These items include
            the items that we persist directly and those referenced by those
            items.
        :param models: All the models that were loaded from database or
            created using the `items`.
        :returns: `creator_keeper` that contains items that can be now created
            because of updated relationships.
            `getter_keeper` that contains items that can be now tried to be
            loaded from database because of updated relationships.
        """
        original_items, original_models = items, models
        
        creator_keeper = ProcessKeeper()
        getter_keeper = ProcessKeeper()
        
        # update_fkey --------------------------------------------------------------
        
        def update_fkey(item_track, fkey, foreign_item_track):
    
            def update_result_keepers():
                item_cls = type(item)
                for keeper, groups in [[creator_keeper[item_cls], item.creators],
                                       [getter_keeper[item_cls], item.getters]]:
                    if groups is None:
                        continue
                    
                    for group in groups:
                        if fkey not in group:
                            continue
                        
                        do_add = True
                        for field in group:
                            if field not in item:
                                do_add = False
                                break
                        if do_add:
                            keeper.append(item_track)
                            break
            
            if foreign_item_track is not None:
                foreign_models = foreign_item_track.models
                if not foreign_models:
                    return
            else:
                foreign_models = [None]
            
            item = item_track.item
            models = item_track.models
            fkeys = item_track.fkeys[fkey]
            
            relation = item.relations[fkey]
            is_x_to_one = relation['relation_type'].is_x_to_one()
            
            if foreign_item_track is not None:
                keepers_updated = False
                for foreign_model in foreign_models:
                    if foreign_model not in fkeys:
                        fkeys.append(foreign_model)
                        if not keepers_updated:
                            update_result_keepers()
            else:
                fkeys.append(None)
    
            for model in models:
                if is_x_to_one:
                    if len(foreign_models) > 1:
                        # cannot reference more then one model
                        raise MultipleModelsMatch(item, fkey, foreign_models)
                    self.set_related_model(item_track.item, model, fkey,
                                           foreign_models[0])
                else:
                    self.add_related_models(item_track.item, model, fkey,
                                            foreign_models)
    
        # saving models to process_keeper ------------------------------------------
        
        for item, models in zip(original_items, original_models):
            if not models:
                continue
            item_cls = type(item)
            
            # getting item_track
            for item_track in self.process_keeper[item_cls]:
                if item_track.item is item:
                    break
                
            for model in models:
                if model not in item_track.models:
                    item_track.models.append(model)
                    
                    if len(item_track.models) > 1 and \
                            not item_cls.allow_multi_update:
                        item_track.models = [
                            item_track.item.resolve_model(item_track.models)
                        ]
                
        # updating relationships ---------------------------------------------------
        
        for item, models in zip(original_items, original_models):
            if not models:
                continue
            item_cls = type(item)
            
            # getting item_track
            for item_track in self.process_keeper[item_cls]:
                if item_track.item is item:
                    break
                   
            # --- forward relations ---
            for f_fkey in item_cls.relations:
                if f_fkey not in item:
                    continue
                
                if item[f_fkey] is None:
                    update_fkey(item_track, f_fkey, None)
                    continue
                
                for f_item in item[f_fkey].as_list():
                    
                    # looking for f_item_track
                    found = False
                    for f_item_cls in self.process_keeper:
                        for f_item_track in self.process_keeper[f_item_cls]:
                            if f_item_track.item is f_item:
                                update_fkey(item_track, f_fkey, f_item_track)
                                found = True
                                break
                        if found:
                            break
    
                    # updating f_fkey
                    
            # --- backward relations ---
            for b_item_cls in self.process_keeper:
                # looking for b_item_track
                for b_item_track in self.process_keeper[b_item_cls]:
                    b_item = b_item_track.item
    
                    for b_fkey, b_value in b_item.data.items():
                        # not an item?
                        if not isinstance(b_value, ItemBase):
                            continue
                         
                        # not referenced?
                        if b_value.is_single_item():
                            if not (b_value is item):
                                continue
                        else:
                            if item not in b_value.bulk:
                                continue
                        
                        # updating b_fkey
                        update_fkey(b_item_track, b_fkey, item_track)
        
        return creator_keeper, getter_keeper


import json
from pprint import pprint
from .utils.mapper import mapper
from .utils.selector import select
from .utils.dict_wrapper import DictWrapper



class ItemBase(object):
    """ This is an abstract item class that serves as a base class for
    single item and bulk item classes.
    
    .. seealso::
        :py:class:`~.item.Item` and :py:class:`~.bulk_item.BulkItem` classes.
    
    :param \*\*kwargs: Values that will be saved as item data.
    """
    
    #--- special methods -------------------------------------------------------
    
    def __init__(self):
        self.data = {}
        mapper.item_created(self)
                
    
    def __setitem__(self, key, value):
        if key in self.data and self.data[key] is value:
            return
        
        old_value = self.data.get(key)
        self.data[key] = value  
        mapper.item_value_set(self, key, old_value, value)
        
    
    def __call__(self, **kwargs):
        for key, value in kwargs.items():
            self[key] = value
        return self
        
    
    def __getitem__(self, key):
        raise NotImplementedError()
    
    
    def __delitem__(self, key):
        old_value = self.data[key]
        del self.data[key]
        mapper.item_value_deleted(self, key, old_value)
    

    def __contains__(self, key):
        raise NotImplementedError()
    
    
    def __copy__(self):
        raise NotImplementedError()
    

    def __deepcopy__(self, memodict={}):
        raise NotImplementedError()
    
    
    #--- utility methods -------------------------------------------------------
         
    def to_dict(self, revert=False):
        """ Converts item into a Python `dict` object.
        
        :param revert: If `True` then converts all not JSON serializable field
            values into types that can be serialized.
        :returns: Python `dict` representation of the item.
        """
        raise NotImplementedError()
    
    
    def dict_wrapper(self):
        """ This method is used for integration with Scrapy project, when
        parsing pages in Scrapy you can yield an item as
        :py:class:`~.utils.dict_wrapper.DictWrapper` (subclass of `dict`)
        and then use :py:meth:`~.utils.dict_wrapper.DictWrapper.get_item`
        method to get the original item.
        
        :returns: An :py:class:`~.utils.dict_wrapper.DictWrapper` class
            instance.
        """
        return DictWrapper(self)
    
    
    def load_dict(self, data):
        """ Loads data from dictionary into the item.
        
        :param data: Dictionary with item data (see :func:`~to_dict` method).
        :returns: The item itself.
        """
        raise NotImplementedError()
    
    
    #--- helper functions ------------------------------------------------------
    
    def pprint(self, as_json=True, revert=None, *args, **kwargs):
        """ Pretty prints the item.
        
        :param as_json: If `True` (default), the a JSON representation of item
            is printed, otherwise dictionary representation is printed using
            `pprint` method from `pprint` module.
        :param revert: Convert all not JSON serializable values into
            serializable ones.
        :param \*args: These arguments passed to `pprint.pprint` function or to
            `json.dumps`.
        :param \**kwargs: These key-value arguments passed to `pprint.pprint`
            function or to `json.dumps`.
        """
        if revert is None:
            revert = as_json
        item_as_dict = self.to_dict(revert=revert)
        
        if not as_json:
            if 'width' not in kwargs:
                kwargs['width'] = 80
            pprint(item_as_dict, *args, **kwargs)
        else:
            json_kwargs = {
                'ensure_ascii': False,
                'indent': 4,
                'sort_keys': True,
            }
            json_kwargs.update(kwargs)
            print(json.dumps(item_as_dict, *args, **json_kwargs))
    
    def select(self, key):
        """ Just calls :py:func:`~.utils.selector.select` function from
        :py:mod:`save_to_db.core.utils.selector` module with `self` as the first
        argument.
        
            .. seealso::
                :py:func:`save_to_db.core.utils.selector.select` function.
        """
        return select(self, key)
    
    
    def get_item_cls(self):
        """ Returns class reference of a single item class that this item
        works with.
        """
        raise NotImplementedError()
    
    
    def is_scoped(self):
        """ Returns `True` if the item class is scoped """
        raise NotImplementedError()
    
    
    def get_collection_id(self):
        """ Returns ID of the collection of the item. """
        raise NotImplementedError()
    
    
    #--- main methods ----------------------------------------------------------
    
    def as_list(self):
        """ Returns a list of items. For single items simply returns a list 
        containing only `self`, for bulk items returns list of all items.  
        """
        raise NotImplementedError()
    
    
    def is_single_item(self):
        """ Returns `True` if item is a single item. """
        raise NotImplementedError()
    
    
    def is_bulk_item(self):
        """ Returns `True` if item is a bulk item. """
        raise NotImplementedError()
    
    
    def revert(self):
        """ Converts all field values into JSON serializable values in such a
        way that :py:meth:`~.process` method converts them bac to original
        values. 
        """
        raise NotImplementedError()
    
    
    def process(self):
        """ Converts all set values to the appropriate data types, sets default
        values if needed, calls :py:meth:`~.process()` method on all referenced
        items.
        
        :returns: Dictionary with single item classes as keys and lists of
            single item instances as values.
        """
        raise NotImplementedError()
    
    
    def delete(self):
        """ Deletes this item. """
        mapper.item_deleted(self)
        
        
        
        
        


from django.db import models
from save_to_db.adapters.utils.relation_type import RelationType


class CommonColumns(models.Model):
    class Meta:
        abstract = True
        
    f_binary = models.BinaryField(null=True)
    f_boolean = models.NullBooleanField(null=True)
    f_string = models.CharField(max_length=255, null=True)
    f_text = models.TextField(null=True)
    f_integer = models.IntegerField(null=True)
    f_float = models.FloatField(null=True)
    f_date = models.DateField(null=True)
    f_time = models.TimeField(null=True)
    f_datetime = models.DateTimeField(null=True)
    

class ModelAutoReverseOne(CommonColumns):
    
    class Meta:
        db_table = 'model_auto_reverse_one'
        
    ITEM_RELATIONS = {
        'two_a_1_1': ('modelautoreverseone', RelationType.ONE_TO_ONE),
        'two_b_1_1': (None, RelationType.ONE_TO_ONE),
        'three_a_x_1': ('modelautoreverseone_set', RelationType.MANY_TO_ONE),
        'three_b_x_1': (None, RelationType.MANY_TO_ONE),
        'four_a_x_x': ('modelautoreverseone_set', RelationType.MANY_TO_MANY),
        'four_b_x_x': (None, RelationType.MANY_TO_MANY),
    }
    
    #--- one-to-one ---
    
    # related name auto configured
    two_a_1_1 = models.OneToOneField(
        'ModelAutoReverseTwoA',
        on_delete=models.CASCADE,
        null=True)
    
    # no related_name
    two_b_1_1 = models.OneToOneField(
        'ModelAutoReverseTwoB',
        on_delete=models.CASCADE,
        related_name='+',
        null=True)
    
    #--- one-to-many ---
    
    # related name auto configured
    three_a_x_1 = models.ForeignKey(
        'ModelAutoReverseThreeA',
        on_delete=models.CASCADE,
        null=True)
    
    # no related_name
    three_b_x_1 = models.ForeignKey(
        'ModelAutoReverseThreeB',
        on_delete=models.CASCADE,
        related_name='+',
        null=True)
    
    #--- many-to-many ---
    
    # related name auto configured
    four_a_x_x = models.ManyToManyField(
        'ModelAutoReverseFourA')
    
    # no related_name
    four_b_x_x = models.ManyToManyField(
        'ModelAutoReverseFourB',
        related_name='+')

class ModelAutoReverseTwoA(CommonColumns):
    class Meta:
        db_table = 'model_auto_reverse_two_a'
    ITEM_RELATIONS = {'modelautoreverseone': ('two_a_1_1',
                                              RelationType.ONE_TO_ONE)}
    
class ModelAutoReverseTwoB(CommonColumns):
    class Meta:
        db_table = 'model_auto_reverse_two_b'
    ITEM_RELATIONS = {}

class ModelAutoReverseThreeA(CommonColumns):
    class Meta:
        db_table = 'model_auto_reverse_thee_a'
    ITEM_RELATIONS = {'modelautoreverseone_set': ('three_a_x_1', 
                                                  RelationType.ONE_TO_MANY)}

class ModelAutoReverseThreeB(CommonColumns):
    class Meta:
        db_table = 'model_auto_reverse_thee_b'
    ITEM_RELATIONS = {}
    
class ModelAutoReverseFourA(CommonColumns):
    class Meta:
        db_table = 'model_auto_reverse_four_a'
    ITEM_RELATIONS = {'modelautoreverseone_set': ('four_a_x_x',
                                                  RelationType.MANY_TO_MANY)}
    
class ModelAutoReverseFourB(CommonColumns):
    class Meta:
        db_table = 'model_auto_reverse_four_b'
    ITEM_RELATIONS = {}


from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestSelect(TestBase):
    
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    def test_select(self):
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
                
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        # single item fild
        item_one = ItemGeneralOne(f_integer=1, f_text='T-1')
        self.assertEqual(item_one.select('f_integer'),
                         [1])
        self.assertEqual(item_one.select('f_text'),
                         ['T-1'])
        self.assertEqual(item_one.select('f_string'),
                         [])
        
        # select from related single item
        item_one = ItemGeneralOne()
        item_one['two_x_1'](f_text='T-1')
        self.assertEqual(item_one.select('two_x_1__f_text'),
                         ['T-1'])
        self.assertEqual(item_one.select('two_x_1__f_string'),
                         [])
        self.assertEqual(item_one.select('two_x_1'),
                         [item_one['two_x_1']])
        
        # select from bulk item
        item_bulk = ItemGeneralOne.Bulk()
        item_bulk.gen(f_integer=1, f_text='T-1')
        item_bulk.gen(f_integer=2)
        item_bulk.gen(f_text='T-3')
        item_bulk.gen(f_float=4.0)
        
        self.assertEqual(item_bulk.select('f_integer'),
                         [1, 2])
        self.assertEqual(item_bulk.select('f_text'),
                         ['T-1', 'T-3'])
        
        # select from many bulks
        top_item_bulk = ItemGeneralOne.Bulk()
        for middle_item_no in range(3):
            middle_item = top_item_bulk.gen(
                f_integer=middle_item_no, f_string='middle-{}'.format(middle_item_no))
            item_bulk = middle_item['two_x_x']
            for item_no in range(2):
                item_bulk.gen(f_integer=middle_item_no, f_float=item_no,
                              f_string='P-{}, I-{}'.format(middle_item_no,
                                                           item_no))
        
        self.assertEqual(top_item_bulk.select('f_string'),
                         ['middle-0', 'middle-1', 'middle-2'])
        self.assertEqual(top_item_bulk.select('two_x_x__f_string'),
                         ['P-0, I-0', 'P-0, I-1',
                          'P-1, I-0', 'P-1, I-1',
                          'P-2, I-0', 'P-2, I-1'])
        self.assertEqual(top_item_bulk.select('two_x_x'),
                         [top_item_bulk.bulk[0]['two_x_x'],
                          top_item_bulk.bulk[1]['two_x_x'],
                          top_item_bulk.bulk[2]['two_x_x']])
        
        
        


ADAPTERS = []  # list of all loaded adapters
"""
When loading an adapter we have to first check if necessary requirements are
met (if particular ORM library is installed), only then we can import
adapter. Otherwise `ImportError` maybe a result of not ORM being not
installed, but actual error in an adapter.
"""

#--- SQL Alchemy ---------------------------------------------------------------
__requirements_installed = False
try:
    import sqlalchemy
    __requirements_installed = True
except ImportError:
    pass

if __requirements_installed:
    from .sqlalchemy_adapter import SqlalchemyAdapter
    ADAPTERS.append(SqlalchemyAdapter)

#--- Django ORM ----------------------------------------------------------------
__requirements_installed = False
try:
    import django
    __requirements_installed = True
except ImportError:
    pass

if __requirements_installed:
    from .django_adapter import DjangoAdapter
    ADAPTERS.append(DjangoAdapter)

#-------------------------------------------------------------------------------





import sqlalchemy as sa
from sqlalchemy.orm import backref, relationship
from save_to_db.adapters.utils.relation_type import RelationType
from ..config import Base


class __CommonColumns(object):
    f_binary = sa.Column(sa.LargeBinary, nullable=True)
    f_boolean = sa.Column(sa.Boolean, nullable=True)
    f_string = sa.Column(sa.String, nullable=True)
    f_text = sa.Column(sa.Text, nullable=True)
    f_integer = sa.Column(sa.Integer, nullable=True)
    f_float = sa.Column(sa.Float, nullable=True)
    f_date = sa.Column(sa.Date, nullable=True)
    f_time = sa.Column(sa.Time, nullable=True)
    f_datetime = sa.Column(sa.DateTime, nullable=True)
    

class ModelAutoReverseOne(__CommonColumns, Base):
    __tablename__ = 'model_auto_reverse_one'
    ITEM_RELATIONS = {
        'two_a_1_1_first': ('one_1_1_first', RelationType.ONE_TO_ONE),
        'two_a_1_1_second': ('one_1_1_second', RelationType.ONE_TO_ONE),
        'two_b_1_1': (None, RelationType.MANY_TO_ONE),  # no reverse setup
        'three_a_x_1_first': ('one_1_x_first', RelationType.MANY_TO_ONE),
        'three_a_x_1_second': ('one_1_x_second', RelationType.MANY_TO_ONE),
        'three_b_x_1': (None, RelationType.MANY_TO_ONE),
        'four_a_x_x_first': ('one_x_x_first', RelationType.MANY_TO_MANY),
        'four_a_x_x_second': ('one_x_x_second', RelationType.MANY_TO_MANY),
        'four_b_x_x': (None, RelationType.MANY_TO_MANY),
    }
    
    id = sa.Column(sa.Integer, primary_key=True)
    
    #--- one-to-one ---
    
    # related name auto configured
    two_a_1_1_first_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_auto_reverse_two_a.id', ondelete='CASCADE'))
    # (this field will be many-to-one, as there is no way to set
    #  `uselist=False` on the other side)
    two_a_1_1_first = relationship(
        'ModelAutoReverseTwoA',
        backref=backref('one_1_1_first', uselist=False),
        foreign_keys=[two_a_1_1_first_id])
    
    # related name manually configured on both sides
    two_a_1_1_second_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_auto_reverse_two_a.id', ondelete='CASCADE'))
    two_a_1_1_second = relationship(
        'ModelAutoReverseTwoA',
        back_populates='one_1_1_second',
        foreign_keys=[two_a_1_1_second_id])
    
    # no related_name (it's a many-to-one-field)
    two_b_1_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_auto_reverse_two_b.id', ondelete='CASCADE'))
    two_b_1_1 = relationship('ModelAutoReverseTwoB')
    
    #--- one-to-many ---
    
    # related name auto configured
    three_a_x_1_first_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_auto_reverse_three_a.id'))
    three_a_x_1_first = relationship(
        'ModelAutoReverseThreeA',
        backref='one_1_x_first',
        foreign_keys=[three_a_x_1_first_id])
    
    # related name manually configured on both sides
    three_a_x_1_second_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_auto_reverse_three_a.id'))
    three_a_x_1_second = relationship(
        'ModelAutoReverseThreeA',
        back_populates='one_1_x_second',
        foreign_keys=[three_a_x_1_second_id])
    
    # no related_name
    three_b_x_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_auto_reverse_three_b.id'))
    three_b_x_1 = relationship('ModelAutoReverseThreeB')
    
    #--- many-to-many ---
    
    # related name auto configured
    four_a_x_x_first = relationship(
        'ModelAutoReverseFourA',
        secondary='model_auto_reverse_one_to_four_a_first',
        backref='one_x_x_first')
    
    # related name manually configured on both sides
    four_a_x_x_second = relationship(
        'ModelAutoReverseFourA',
        secondary='model_auto_reverse_one_to_four_a_second',
        back_populates='one_x_x_second')
    
    # no related_name
    
    four_b_x_x = relationship(
        'ModelAutoReverseFourB',
        secondary='model_auto_reverse_one_to_four_b')
    

class ModelAutoReverseTwoA(__CommonColumns, Base):
    __tablename__ = 'model_auto_reverse_two_a'
    ITEM_RELATIONS = {
        'one_1_1_first': ('two_a_1_1_first', RelationType.ONE_TO_ONE),
        'one_1_1_second': ('two_a_1_1_second', RelationType.ONE_TO_ONE),
    }
    
    id = sa.Column(sa.Integer, primary_key=True)
    one_1_1_second = relationship(
        ModelAutoReverseOne,
        back_populates='two_a_1_1_second',
        foreign_keys=[ModelAutoReverseOne.two_a_1_1_second_id],
        uselist=False)
    

class ModelAutoReverseTwoB(__CommonColumns, Base):
    __tablename__ = 'model_auto_reverse_two_b'
    ITEM_RELATIONS = {}
    
    id = sa.Column(sa.Integer, primary_key=True)
    
    
class ModelAutoReverseThreeA(__CommonColumns, Base):
    __tablename__ = 'model_auto_reverse_three_a'
    ITEM_RELATIONS = {
        'one_1_x_first': ('three_a_x_1_first', RelationType.ONE_TO_MANY),
        'one_1_x_second': ('three_a_x_1_second', RelationType.ONE_TO_MANY),
    }

    
    id = sa.Column(sa.Integer, primary_key=True)
    one_1_x_second = relationship(
        ModelAutoReverseOne,
        back_populates='three_a_x_1_second',
        foreign_keys=[ModelAutoReverseOne.three_a_x_1_second_id])
    

class ModelAutoReverseThreeB(__CommonColumns, Base):
    __tablename__ = 'model_auto_reverse_three_b'
    ITEM_RELATIONS = {}
    
    id = sa.Column(sa.Integer, primary_key=True)
    
    
class ModelAutoReverseFourA(__CommonColumns, Base):
    __tablename__ = 'model_auto_reverse_four_a'
    ITEM_RELATIONS = {
        'one_x_x_first': ('four_a_x_x_first', RelationType.MANY_TO_MANY),
        'one_x_x_second': ('four_a_x_x_second', RelationType.MANY_TO_MANY),
    }
    
    id = sa.Column(sa.Integer, primary_key=True)
    
    one_x_x_second = relationship(
        ModelAutoReverseOne,
        secondary='model_auto_reverse_one_to_four_a_second',
        back_populates='four_a_x_x_second')


class ModelAutoReverseOneToFourAFirst(Base):
    __tablename__ = 'model_auto_reverse_one_to_four_a_first'
    
    one_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(ModelAutoReverseOne.id, ondelete='CASCADE'),
        primary_key=True)
    one = relationship(
        ModelAutoReverseOne,
        foreign_keys=[one_id])
    
    four_a_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(ModelAutoReverseFourA.id, ondelete='CASCADE'),
        primary_key=True)
    four_a = relationship(
        ModelAutoReverseFourA,
        foreign_keys=[four_a_id])


class ModelAutoReverseOneToFourASecond(Base):
    __tablename__ = 'model_auto_reverse_one_to_four_a_second'
    
    one_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(ModelAutoReverseOne.id, ondelete='CASCADE'),
        primary_key=True)
    one = relationship(
        ModelAutoReverseOne,
        foreign_keys=[one_id])
    
    four_a_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(ModelAutoReverseFourA.id, ondelete='CASCADE'),
        primary_key=True)
    four_a = relationship(
        ModelAutoReverseFourA,
        foreign_keys=[four_a_id])
    

class ModelAutoReverseFourB(__CommonColumns, Base):
    __tablename__ = 'model_auto_reverse_four_b'
    
    id = sa.Column(sa.Integer, primary_key=True)
    ITEM_RELATIONS = {}
    

class ModelAutoReverseOneToFourB(Base):
    __tablename__ = 'model_auto_reverse_one_to_four_b'
    
    one_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(ModelAutoReverseOne.id, ondelete='CASCADE'),
        primary_key=True)
    one = relationship(
        ModelAutoReverseOne,
        foreign_keys=[one_id])
    
    four_b_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(ModelAutoReverseFourB.id, ondelete='CASCADE'),
        primary_key=True)
    four_b = relationship(
        ModelAutoReverseFourB,
        foreign_keys=[four_b_id])


# Generated by Django 2.1.4 on 2019-05-10 12:59

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ModelAutoReverseFourA',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
            ],
            options={
                'db_table': 'model_auto_reverse_four_a',
            },
        ),
        migrations.CreateModel(
            name='ModelAutoReverseFourB',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
            ],
            options={
                'db_table': 'model_auto_reverse_four_b',
            },
        ),
        migrations.CreateModel(
            name='ModelAutoReverseOne',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
                ('four_a_x_x', models.ManyToManyField(to='orm_only.ModelAutoReverseFourA')),
                ('four_b_x_x', models.ManyToManyField(related_name='_modelautoreverseone_four_b_x_x_+', to='orm_only.ModelAutoReverseFourB')),
            ],
            options={
                'db_table': 'model_auto_reverse_one',
            },
        ),
        migrations.CreateModel(
            name='ModelAutoReverseThreeA',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
            ],
            options={
                'db_table': 'model_auto_reverse_thee_a',
            },
        ),
        migrations.CreateModel(
            name='ModelAutoReverseThreeB',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
            ],
            options={
                'db_table': 'model_auto_reverse_thee_b',
            },
        ),
        migrations.CreateModel(
            name='ModelAutoReverseTwoA',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
            ],
            options={
                'db_table': 'model_auto_reverse_two_a',
            },
        ),
        migrations.CreateModel(
            name='ModelAutoReverseTwoB',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
            ],
            options={
                'db_table': 'model_auto_reverse_two_b',
            },
        ),
        migrations.CreateModel(
            name='ModelConstraintsFive',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
            ],
            options={
                'db_table': 'model_constraints_five',
            },
        ),
        migrations.CreateModel(
            name='ModelConstraintsFour',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
            ],
            options={
                'db_table': 'model_constraints_four',
            },
        ),
        migrations.CreateModel(
            name='ModelConstraintsOne',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
                ('f_text', models.TextField()),
                ('f_integer', models.IntegerField(null=True, unique=True)),
                ('f_string', models.CharField(max_length=255, unique=True)),
            ],
            options={
                'db_table': 'model_constraints_one',
            },
        ),
        migrations.CreateModel(
            name='ModelConstraintsSelf',
            fields=[
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
                ('code', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('first_parent_1_1', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='first_child_1_1', to='orm_only.ModelConstraintsSelf')),
                ('parent_x_1', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='child_1_x', to='orm_only.ModelConstraintsSelf')),
                ('parent_x_x', models.ManyToManyField(related_name='child_x_x', to='orm_only.ModelConstraintsSelf')),
                ('second_parent_1_1', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='second_child_1_1', to='orm_only.ModelConstraintsSelf')),
            ],
            options={
                'db_table': 'model_constraints_self',
            },
        ),
        migrations.CreateModel(
            name='ModelConstraintsSix',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
                ('five_1_1', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='six_1_1', to='orm_only.ModelConstraintsFive')),
            ],
            options={
                'db_table': 'model_constraints_six',
            },
        ),
        migrations.CreateModel(
            name='ModelConstraintsThree',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
                ('four_x_1', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='three_1_x', to='orm_only.ModelConstraintsFour')),
                ('one_x_1', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='three_1_x', to='orm_only.ModelConstraintsOne')),
            ],
            options={
                'db_table': 'model_constraints_three',
            },
        ),
        migrations.CreateModel(
            name='ModelConstraintsTwo',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
                ('four_x_1', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='two_1_x', to='orm_only.ModelConstraintsFour')),
                ('one_x_1', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='two_1_x', to='orm_only.ModelConstraintsOne')),
            ],
            options={
                'db_table': 'model_constraints_two',
            },
        ),
        migrations.CreateModel(
            name='ModelFieldTypes',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('binary_1', models.BinaryField(db_column='new_binary_1')),
                ('string_1', models.CharField(db_column='new_string_1', max_length=255)),
                ('text_1', models.TextField(db_column='new_text_1')),
                ('integer_1', models.IntegerField(db_column='new_integer_1')),
                ('integer_2', models.BigIntegerField(db_column='new_integer_2')),
                ('integer_3', models.PositiveIntegerField(db_column='new_integer_3')),
                ('integer_4', models.PositiveSmallIntegerField(db_column='new_integer_4')),
                ('integer_5', models.SmallIntegerField(db_column='new_integer_5')),
                ('boolean_1', models.BooleanField(db_column='new_boolean_1')),
                ('boolean_2', models.NullBooleanField(db_column='new_boolean_2')),
                ('float_1', models.FloatField(db_column='new_float_1')),
                ('decimal_1', models.DecimalField(db_column='new_decimal_2', decimal_places=3, max_digits=12)),
                ('date_1', models.DateField(db_column='new_date_1')),
                ('time_1', models.TimeField(db_column='new_time_1')),
                ('datetime_1', models.DateTimeField(db_column='new_datetime_1')),
            ],
            options={
                'db_table': 'model_field_types',
            },
        ),
        migrations.CreateModel(
            name='ModelGeneralOne',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_decimal', models.DecimalField(decimal_places=6, max_digits=18, null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
                ('parent_1_1', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='child_1_1', to='orm_only.ModelGeneralOne')),
                ('parent_x_1', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='child_1_x', to='orm_only.ModelGeneralOne')),
                ('parent_x_x', models.ManyToManyField(related_name='child_x_x', to='orm_only.ModelGeneralOne')),
            ],
            options={
                'db_table': 'model_general_one',
            },
        ),
        migrations.CreateModel(
            name='ModelGeneralTwo',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_decimal', models.DecimalField(decimal_places=6, max_digits=18, null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
                ('one_1_1', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='two_1_1', to='orm_only.ModelGeneralOne')),
                ('one_x_1', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='two_1_x', to='orm_only.ModelGeneralOne')),
                ('parent_1_1', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='child_1_1', to='orm_only.ModelGeneralTwo')),
                ('parent_x_1', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='child_1_x', to='orm_only.ModelGeneralTwo')),
                ('parent_x_x', models.ManyToManyField(related_name='child_x_x', to='orm_only.ModelGeneralTwo')),
            ],
            options={
                'db_table': 'model_general_two',
            },
        ),
        migrations.CreateModel(
            name='ModelManyRefsOne',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
            ],
            options={
                'db_table': 'model_many_refs_one',
            },
        ),
        migrations.CreateModel(
            name='ModelManyRefsTwo',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
                ('one_1_1_b', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='two_1_1_b', to='orm_only.ModelManyRefsOne')),
                ('one_x_1_b', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='two_1_x_b', to='orm_only.ModelManyRefsOne')),
                ('one_x_x_b', models.ManyToManyField(related_name='two_x_x_b', to='orm_only.ModelManyRefsOne')),
            ],
            options={
                'db_table': 'model_many_refs_two',
            },
        ),
        migrations.CreateModel(
            name='ModelMergeOne',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
            ],
            options={
                'db_table': 'model_merge_one',
            },
        ),
        migrations.CreateModel(
            name='ModelMergeOneX',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
            ],
            options={
                'db_table': 'model_merge_one_x',
            },
        ),
        migrations.CreateModel(
            name='ModelMergeThree',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
            ],
            options={
                'db_table': 'model_merge_three',
            },
        ),
        migrations.CreateModel(
            name='ModelMergeThreeX',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
            ],
            options={
                'db_table': 'model_merge_three_x',
            },
        ),
        migrations.CreateModel(
            name='ModelMergeTwo',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
                ('three_1_1', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='two_1_1', to='orm_only.ModelMergeThree')),
                ('three_x_1', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='two_1_x', to='orm_only.ModelMergeThree')),
            ],
            options={
                'db_table': 'model_merge_two',
            },
        ),
        migrations.CreateModel(
            name='ModelMergeTwoX',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('f_binary', models.BinaryField(null=True)),
                ('f_boolean', models.NullBooleanField()),
                ('f_string', models.CharField(max_length=255, null=True)),
                ('f_text', models.TextField(null=True)),
                ('f_integer', models.IntegerField(null=True)),
                ('f_float', models.FloatField(null=True)),
                ('f_date', models.DateField(null=True)),
                ('f_time', models.TimeField(null=True)),
                ('f_datetime', models.DateTimeField(null=True)),
                ('three_x_1', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='two_1_x', to='orm_only.ModelMergeThreeX')),
            ],
            options={
                'db_table': 'model_merge_two_x',
            },
        ),
        migrations.AddField(
            model_name='modelmergeonex',
            name='two_x_1',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='one_1_x', to='orm_only.ModelMergeTwoX'),
        ),
        migrations.AddField(
            model_name='modelmergeone',
            name='two_1_1',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='one_1_1', to='orm_only.ModelMergeTwo'),
        ),
        migrations.AddField(
            model_name='modelmergeone',
            name='two_x_1',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='one_1_x', to='orm_only.ModelMergeTwo'),
        ),
        migrations.AddField(
            model_name='modelmanyrefsone',
            name='two_1_1_a',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='one_1_1_a', to='orm_only.ModelManyRefsTwo'),
        ),
        migrations.AddField(
            model_name='modelmanyrefsone',
            name='two_x_1_a',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='one_1_x_a', to='orm_only.ModelManyRefsTwo'),
        ),
        migrations.AddField(
            model_name='modelmanyrefsone',
            name='two_x_x_a',
            field=models.ManyToManyField(related_name='one_x_x_a', to='orm_only.ModelManyRefsTwo'),
        ),
        migrations.AddField(
            model_name='modelgeneralone',
            name='two_x_1',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='one_1_x', to='orm_only.ModelGeneralTwo'),
        ),
        migrations.AddField(
            model_name='modelgeneralone',
            name='two_x_x',
            field=models.ManyToManyField(related_name='one_x_x', to='orm_only.ModelGeneralTwo'),
        ),
        migrations.AddField(
            model_name='modelconstraintsthree',
            name='two_x_x',
            field=models.ManyToManyField(related_name='three_x_x', to='orm_only.ModelConstraintsTwo'),
        ),
        migrations.AlterUniqueTogether(
            name='modelconstraintsfour',
            unique_together={('f_integer', 'f_string')},
        ),
        migrations.AddField(
            model_name='modelconstraintsfive',
            name='four_1_1',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='five_1_1', to='orm_only.ModelConstraintsFour'),
        ),
        migrations.AddField(
            model_name='modelconstraintsfive',
            name='one_1_1',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='five_1_1', to='orm_only.ModelConstraintsOne'),
        ),
        migrations.AddField(
            model_name='modelautoreverseone',
            name='three_a_x_1',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='orm_only.ModelAutoReverseThreeA'),
        ),
        migrations.AddField(
            model_name='modelautoreverseone',
            name='three_b_x_1',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='orm_only.ModelAutoReverseThreeB'),
        ),
        migrations.AddField(
            model_name='modelautoreverseone',
            name='two_a_1_1',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, to='orm_only.ModelAutoReverseTwoA'),
        ),
        migrations.AddField(
            model_name='modelautoreverseone',
            name='two_b_1_1',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='orm_only.ModelAutoReverseTwoB'),
        ),
        migrations.AlterUniqueTogether(
            name='modelmergetwox',
            unique_together={('f_integer', 'three_x_1')},
        ),
        migrations.AlterUniqueTogether(
            name='modelmergeonex',
            unique_together={('f_integer', 'two_x_1')},
        ),
        migrations.AlterUniqueTogether(
            name='modelconstraintssix',
            unique_together={('five_1_1', 'f_integer')},
        ),
    ]



from .adapter_except import *
from .item_collection import *
from .item_config import *
from .item_field import *
from .item_persist import *
from .item_process import *
from .merge_policy import *
from .scope_except import *



class Signal(object):
    """ This class is used to emit signals.
    
    :ivar name: Signal name.
    """
    
    def __init__(self, name):
        self.name = name
        self.__listeners = []
    
    def register(self, listener):
        """ Registers listener to the signal.
        
            .. note::
                This method can be also used as a decorator:
                
                .. code-block:: Python
                
                    signal = Signal('MySignal')
                    
                    @signal.register
                    def signal_listener(value):
                        print('Signal value: {}'.format(value))
                    
                    # Output: "Signal value: Demo Value"
                    signal.emit('Demo Value')
        """
        if listener not in self.__listeners:
            self.__listeners.append(listener)
        # returning listener makes it possible to use this function as
        # a decorator
        return listener
    
    def unregister(self, listener):
        """ Unregisters listener from the signal. """
        if listener in self.__listeners:
            self.__listeners.remove(listener)
    
    def clear(self):
        """ Removes all listeners from the signal. """
        self.__listeners.clear()
    
    def emit(self, *args, **kwargs):
        """ Emits the signal. All listeners will be called with the same
        arguments as this method.
        """
        for listener in self.__listeners:
            listener(*args, **kwargs)
    
    def __repr__(self):
        return '<Signal(\'{}\')>'.format(self.name)


import datetime
from decimal import Decimal

from io import BytesIO
from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestItemDump(TestBase):

    ModelGeneralOne = None
    ModelGeneralTwo = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        
        class ItemFieldTypes(Item):
            model_cls = cls.ModelFieldTypes
        cls.ItemFieldTypes = ItemFieldTypes
        
        class ItemGeneralOne(Item):
            model_cls = cls.ModelGeneralOne
        cls.ItemGeneralOne = ItemGeneralOne
        
        class ItemGeneralTwo(Item):
            model_cls = cls.ModelGeneralTwo
        cls.ItemGeneralTwo = ItemGeneralTwo
    
    
    def check_dumps(self, item, expected):
        # converting to bytes
        item_as_bytes = self.persister.dumps(item)
        self.assertIsInstance(item_as_bytes, bytes)
        
        item = self.persister.loads(item_as_bytes)    
        self.assertEqual(item.to_dict(), expected)
        
        # saving to a file-like object        
        storage = BytesIO()
        self.persister.dump(item, storage)
    
        storage.seek(0)
        item = self.persister.load(storage)
        self.assertEqual(item.to_dict(), expected)
    
    
    def test_dump_types(self):
        item = self.ItemGeneralOne(f_binary='binary data',
                                   f_boolean='true',
                                   f_string='string data',
                                   f_text='text_data',
                                   f_integer='10',
                                   f_float='10.10',
                                   f_date='2000-10-30',
                                   f_time='20:30:40',
                                   f_datetime='2000-10-30 20:30:40')
        expected = {
            'item': {
                'f_binary': b'binary data',
                'f_boolean': True,
                'f_date': datetime.date(2000, 10, 30),
                'f_datetime': datetime.datetime(2000, 10, 30, 20, 30, 40),
                'f_float': 10.1,
                'f_integer': 10,
                'f_string': 'string data',
                'f_text': 'text_data',
                'f_time': datetime.time(20, 30, 40),
            }
        }
        self.check_dumps(item, expected)


    def test_dump_single_item_with_relations(self):
        def gen_item():
            item = self.ItemGeneralOne(f_integer='1')
            item['two_1_1__f_integer'] = '11'
            item['two_x_x'].gen(f_float='99.1')
            item['two_x_x'].gen(f_float='99.2', one_x_1__f_integer='192') 
            item['two_x_x__f_float'] = '1000.1'
            return item
        
        item = gen_item()
        item.process()
        expect = item.to_dict()
        self.check_dumps(gen_item(), expect)
        
        
    def test_dump_bulk_item_with_relations(self):
        def gen_item():
            bulk = self.ItemGeneralOne.Bulk()
            bulk.gen(f_float='1.1')
            bulk.gen(f_float='1.2', two_x_1__f_integer='2') 
            bulk['two_x_x__f_float'] = '1000.1'
            bulk['two_x_1'] = self.ItemGeneralTwo(f_boolean='true')
            return bulk

        item = gen_item()
        item.process()
        expect = item.to_dict()
        self.check_dumps(gen_item(), expect)
    
    
    def test_multiple_items_dump(self):
        storage = BytesIO()
        
        self.persister.dump(self.ItemGeneralOne(f_integer='1'), storage)
        self.persister.dump(self.ItemGeneralOne(f_integer='2'), storage)
        self.persister.dump(self.ItemGeneralOne(f_integer='3'), storage)
        
        storage.seek(0)
        
        dict_wrapper = self.persister.load(storage).to_dict()
        self.assertEqual(dict_wrapper, {'item': {'f_integer': 1}})
        
        dict_wrapper = self.persister.load(storage).to_dict()
        self.assertEqual(dict_wrapper, {'item': {'f_integer': 2}})
        
        dict_wrapper = self.persister.load(storage).to_dict()
        self.assertEqual(dict_wrapper, {'item': {'f_integer': 3}})
        
        # no more items
        dict_wrapper = self.persister.load(storage)
        self.assertIsNone(dict_wrapper)


    def test_arbitrary_id_position(self):
        # normal item
        item_one = self.ItemGeneralOne(f_integer='1')
        child_item = item_one['child_1_x'].gen(f_integer='2')
        item_one['child_x_x'].add(child_item)
        item_one.process()
        expect = {
            'id': 1,
            'item': {
                'child_1_x': {
                    'bulk': [
                        {
                            'id': 2,
                            'item': {
                                'f_integer': 2,
                                'parent_x_1': {
                                    'id': 1
                                },
                                'parent_x_x': {
                                    'bulk': [
                                        {
                                            'id': 1
                                        }
                                    ],
                                    'defaults': {}
                                }
                            }
                        }
                    ],
                    'defaults': {}
                },
                'child_x_x': {
                    'bulk': [
                        {
                            'id': 2
                        }
                    ],
                    'defaults': {}
                },
                'f_integer': 1
            }
        }
        loaded_item = self.ItemGeneralOne().load_dict(expect)
        self.assertEqual(loaded_item.to_dict(), expect)
        
        # moving parent data
        item_as_dict = {
            'id': 1,
            'item': {
                'child_1_x': {
                    'bulk': [
                        {
                            'id': 2
                        }
                    ],
                    'defaults': {}
                },
                'child_x_x': {
                    'bulk': [
                        {
                            'id': 2,
                            'item': {
                                'f_integer': 2,
                                'parent_x_1': {
                                    'id': 1
                                },
                                'parent_x_x': {
                                    'bulk': [
                                        {
                                            'id': 1
                                        }
                                    ],
                                    'defaults': {}
                                }
                            }
                        }
                    ],
                    'defaults': {}
                },
                'f_integer': 1
            }
        }
        loaded_item = self.ItemGeneralOne().load_dict(item_as_dict)
        self.assertEqual(loaded_item.to_dict(), expect)
        
        # id in default
        item_one = self.ItemGeneralOne.Bulk()
        item_one.gen(f_integer='1')
        other_item = item_one.gen(f_integer='2')
        item_one['parent_x_1'] = other_item
        item_one.process()
        expect = {
            'bulk': [
                {
                    'id': 2,
                    'item': {
                        'f_integer': 1,
                        'parent_x_1': {
                            'item': {
                                'child_1_x': {
                                    'bulk': [
                                        {
                                            'id': 2
                                        }
                                    ],
                                    'defaults': {}
                                },
                                'f_integer': 2
                            }
                        }
                    }
                },
                {
                    'id': 1
                }
            ],
            'defaults': {
                'parent_x_1': {
                    'id': 1,
                    'item': {
                        'f_integer': 2,
                        'parent_x_1': {
                            'item': {
                                'child_1_x': {
                                    'bulk': [
                                        {
                                            'id': 1
                                        }
                                    ],
                                    'defaults': {}
                                },
                                'f_integer': 2
                            }
                        }
                    }
                }
            }
        }
        loaded_item = self.ItemGeneralOne.Bulk().load_dict(expect)
        self.assertEqual(loaded_item.to_dict(), expect)
        
        # moving other item data
        item_as_dict = {
            'bulk': [
                {
                    'id': 2,
                    'item': {
                        'f_integer': 1,
                        'parent_x_1': {
                            'item': {
                                'child_1_x': {
                                    'bulk': [
                                        {
                                            'id': 2
                                        }
                                    ],
                                    'defaults': {}
                                },
                                'f_integer': 2
                            }
                        }
                    }
                },
                {
                    'id': 1,
                    'item': {
                        'f_integer': 2,
                        'parent_x_1': {
                            'item': {
                                'child_1_x': {
                                    'bulk': [
                                        {
                                            'id': 1
                                        }
                                    ],
                                    'defaults': {}
                                },
                                'f_integer': 2
                            }
                        }
                    }
                }
            ],
            'defaults': {
                'parent_x_1': {
                    'id': 1,
                  }
            }
        }
        loaded_item = self.ItemGeneralOne.Bulk().load_dict(item_as_dict)
        self.assertEqual(loaded_item.to_dict(), expect)
    
    
    def test_dump_instance_config(self):
        self.item_cls_manager.clear()
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            update_only_mode = True
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            get_only_mode = True
            
        item_one_bulk = ItemGeneralOne.Bulk()
        item_one_bulk.gen(f_integer=1)
        
        item_one = item_one_bulk.gen(f_integer=2)
        item_one.get_only_mode = True
        item_one.update_only_mode = False
        
        
        item_one['two_1_x'].gen(f_integer=201)
        
        item_two = item_one['two_1_x'].gen(f_integer=202)
        item_two.get_only_mode = False
        item_two.update_only_mode = True
        
        expected = {
            'bulk': [
                {
                    'item': {
                        'f_integer': 1
                    }
                },
                {
                    'id': 1,
                    'item': {
                        'f_integer': 2,
                        'two_1_x': {
                            'bulk': [
                                {
                                    'item': {
                                        'f_integer': 201,
                                        'one_x_1': {
                                            'id': 1
                                        }
                                    }
                                },
                                {
                                    'item': {
                                        'f_integer': 202,
                                        'one_x_1': {
                                            'id': 1
                                        }
                                    },
                                    # overwritten values
                                    'get_only_mode': False,
                                    'update_only_mode': True
                                }
                            ],
                            'defaults': {}
                        }
                    },
                    # overwritten values
                    'get_only_mode': True,
                    'update_only_mode': False
                }
            ],
            'defaults': {}
        }

        self.check_dumps(item_one_bulk, expected)


    def test_revert_and_to_dict(self):
        self.item_cls_manager.clear()
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            conversions = {
                'date_formats': '%d.%m.%Y',
                'time_formats': '%H-%M-%S',
                'datetime_formats': '%d.%m.%Y %H-%M-%S',
                'decimal_separator': '!',
            }
         
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        # from single item -----------------------------------------------------
        one = ItemGeneralOne(f_integer='10',
                             f_decimal='10!10',
                             f_date='20.10.2010',
                             f_time='10-20-30',
                             f_datetime='20.10.2010 10-20-30')
        one['two_1_1'] = ItemGeneralTwo(f_integer='20',
                                        f_decimal='20.20',
                                        f_date='2010-10-20',
                                        f_time='10:20:30',
                                        f_datetime='2010-10-20 10:20:30')
        one.process()
        
        expected_fields = {
            'f_integer': 10,
            'f_decimal': Decimal('10.10'),
            'f_date': datetime.date(2010, 10, 20),
            'f_time': datetime.time(10, 20, 30),
            'f_datetime': datetime.datetime(2010, 10, 20, 10, 20, 30),
            
            'two_1_1__f_integer': 20,
            'two_1_1__f_decimal': Decimal('20.20'),
            'two_1_1__f_date': datetime.date(2010, 10, 20),
            'two_1_1__f_time': datetime.time(10, 20, 30),
            'two_1_1__f_datetime': datetime.datetime(2010, 10, 20, 10, 20, 30),
        }
        for key, value in expected_fields.items():
            self.assertIs(type(one[key]), type(value))
            self.assertEqual(one[key], value)

        
        expect_reverted = {
            'f_integer': 10,
            'f_decimal': '10!10',
            'f_date': '20.10.2010',
            'f_time': '10-20-30',
            'f_datetime': '20.10.2010 10-20-30',
            
            'two_1_1__f_integer': 20,
            'two_1_1__f_decimal': '20.20',
            'two_1_1__f_date': '2010-10-20',
            'two_1_1__f_time': '10:20:30',
            'two_1_1__f_datetime': '2010-10-20 10:20:30',
        }
        reverted_dict = one.to_dict(revert=True)

        for key, value in expect_reverted.items():
            if '__' not in key:
                self.assertEqual(reverted_dict['item'][key], value)
            else:
                key = key.rsplit('__', 1)[-1]
                self.assertEqual(reverted_dict['item']['two_1_1']['item'][key],
                                 value)
                
        # original not changed
        original_dict = one.to_dict()

        for key, value in expected_fields.items():
            if '__' not in key:
                self.assertEqual(original_dict['item'][key], value)
            else:
                key = key.rsplit('__', 1)[-1]
                self.assertEqual(original_dict['item']['two_1_1']['item'][key],
                                 value)
        
        # now all reverted
        one.revert()
        for key, value in expect_reverted.items():
            self.assertIs(type(one[key]), type(value))
            self.assertEqual(one[key], value)
        
        # again normal
        one.process()
        for key, value in expected_fields.items():
            self.assertIs(type(one[key]), type(value))
            self.assertEqual(one[key], value)
        

        # from bulk item -------------------------------------------------------
        bulk_one = ItemGeneralOne.Bulk(f_decimal='1000!1000')
        
        # bulk
        bulk_one.gen(f_integer=10,
                     f_decimal='10!10',
                     f_date='20.10.2010',
                     f_time='10-20-30',
                     f_datetime='20.10.2010 10-20-30')
        
        for _ in range(2):
            bulk_one['two_x_x'].gen(f_integer=20,
                                    f_decimal='20.20',
                                    f_date='2010-10-20',
                                    f_time='10:20:30',
                                    f_datetime='2010-10-20 10:20:30')

        bulk_one.process()

        expect_processed = {
            'bulk': [{
                'id': 2
            }],
            'defaults': {
                'f_decimal': Decimal('1000.1000'),
                'two_x_x': {
                    'bulk': [{
                        'item': {
                            'f_date': datetime.date(2010, 10, 20),
                            'f_datetime': datetime.datetime(
                                2010, 10, 20,10, 20, 30
                            ),
                            'f_decimal': Decimal('20.20'),
                            'f_integer': 20,
                            'f_time': datetime.time(10, 20, 30),
                            'one_x_x': {
                                'bulk': [{
                                    'id': 2,
                                    'item': {
                                        'f_date': datetime.date(2010, 10, 20),
                                        'f_datetime': datetime.datetime(
                                            2010, 10, 20, 10, 20, 30
                                        ),
                                        'f_decimal': Decimal('10.10'),
                                        'f_integer': 10,
                                        'f_time': datetime.time(10, 20, 30),
                                        'two_x_x': {
                                            'id': 1
                                        }
                                    }
                                }],
                                'defaults': {}
                            }
                        }
                    }, {
                        'item': {
                            'f_date': datetime.date(2010, 10, 20),
                            'f_datetime': datetime.datetime(
                                2010, 10, 20, 10, 20, 30
                            ),
                            'f_decimal': Decimal('20.20'),
                            'f_integer': 20,
                            'f_time': datetime.time(10, 20, 30),
                            'one_x_x': {
                                'bulk': [{
                                    'id': 2
                                }],
                                'defaults': {}
                            }
                        }
                    }],
                    'defaults': {},
                    'id': 1
                }
            }
        }
        
        self.assertEqual(bulk_one.to_dict(), expect_processed)
        
        expect_reverted = {
            'bulk': [{
                'id': 2
            }],
            'defaults': {
                'two_x_x': {
                    'id': 1,
                    'bulk': [{
                        'item': {
                            'f_datetime': '2010-10-20 10:20:30',
                            'f_integer': 20,
                            'f_date': '2010-10-20',
                            'f_decimal': '20.20',
                            'one_x_x': {
                                'bulk': [{
                                    'id': 2,
                                    'item': {
                                        'f_datetime': '20.10.2010 10-20-30',
                                        'f_integer': 10,
                                        'two_x_x': {
                                            'id': 1
                                        },
                                        'f_date': '20.10.2010',
                                        'f_decimal': '10!10',
                                        'f_time': '10-20-30'
                                    }
                                }],
                                'defaults': {}
                            },
                            'f_time': '10:20:30'
                        }
                    }, {
                        'item': {
                            'f_datetime': '2010-10-20 10:20:30',
                            'f_integer': 20,
                            'f_date': '2010-10-20',
                            'f_decimal': '20.20',
                            'one_x_x': {
                                'bulk': [{
                                    'id': 2
                                }],
                                'defaults': {}
                            },
                            'f_time': '10:20:30'
                        }
                    }],
                    'defaults': {}
                },
                'f_decimal': '1000!1000'
            }
        }
        
        self.assertEqual(bulk_one.to_dict(revert=True), expect_reverted)
        # original not changed
        self.assertEqual(bulk_one.to_dict(), expect_processed)
        
        # all reverted
        bulk_one.revert()
        self.assertEqual(bulk_one.to_dict(revert=True), expect_reverted)
        self.assertEqual(bulk_one.to_dict(), expect_reverted)
        
        # back again
        bulk_one.process()
        self.assertEqual(bulk_one.to_dict(), expect_processed)
        self.assertEqual(bulk_one.to_dict(revert=True), expect_reverted)
    
    
    def test_multiple_date_time_formats_to_dict_revert(self):
        self.item_cls_manager.clear()
        self.item_cls_manager.autogenerate = True
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            conversions = {
                'date_formats': ['%d.%m.%Y', '%d.%m.%y'],
                'time_formats': ['%H:%M:%S', '%H:%M'],
                'datetime_formats': ['%d.%m.%Y %H:%M:%S', '%d.%m.%y %H:%M'],
            }
        
        # long format versions
        first_item = ItemGeneralOne(f_date='20.10.2010',
                                    f_time='10:20:30',
                                    f_datetime='20.10.2010 10:20:30')
        first_item.process()
        expect = {
            'item': {
                'f_time': datetime.time(10, 20, 30),
                'f_date': datetime.date(2010, 10, 20),
                'f_datetime': datetime.datetime(2010, 10, 20, 10, 20, 30),
            }
        }
        self.assertEqual(first_item.to_dict(), expect)
        
        first_item.revert()
        expect = {
            'item': {
                'f_time': '10:20:30',
                'f_datetime': '20.10.2010 10:20:30',
                'f_date': '20.10.2010',
            }
        }
        self.assertEqual(first_item.to_dict(), expect)
        
        # short format versions
        first_item = ItemGeneralOne(f_date='20.10.10',
                                    f_time='10:20',
                                    f_datetime='20.10.10 10:20')
        first_item.process()
        expect = {
            'item': {
                'f_time': datetime.time(10, 20, 0),
                'f_date': datetime.date(2010, 10, 20),
                'f_datetime': datetime.datetime(2010, 10, 20, 10, 20, 0),
            }
        }
        self.assertEqual(first_item.to_dict(), expect)
        
        first_item.revert()  # long versions will be used
        expect = {
            'item': {
                'f_time': '10:20:00',
                'f_datetime': '20.10.2010 10:20:00',
                'f_date': '20.10.2010',
            }
        }
        self.assertEqual(first_item.to_dict(), expect)

            


from django.db import models


class CommonColumns(models.Model):
    class Meta:
        abstract = True
        
    f_binary = models.BinaryField(null=True)
    f_boolean = models.NullBooleanField(null=True)
    f_string = models.CharField(max_length=255, null=True)
    f_text = models.TextField(null=True)
    f_integer = models.IntegerField(null=True)
    f_float = models.FloatField(null=True)
    f_date = models.DateField(null=True)
    f_time = models.TimeField(null=True)
    f_datetime = models.DateTimeField(null=True)
    

class ModelManyRefsOne(CommonColumns):
    class Meta:
        db_table = 'model_many_refs_one'
    
    two_x_1_a = models.ForeignKey(
        'ModelManyRefsTwo',
        on_delete=models.CASCADE,
        related_name='one_1_x_a',
        null=True)
    
    two_x_x_a = models.ManyToManyField(
        'ModelManyRefsTwo',
        related_name='one_x_x_a')
    
    two_1_1_a = models.OneToOneField(
        'ModelManyRefsTwo',
        on_delete=models.CASCADE,
        related_name='one_1_1_a',
        null=True)


class ModelManyRefsTwo(CommonColumns):
    class Meta:
        db_table = 'model_many_refs_two'
    
    one_x_1_b = models.ForeignKey(
        'ModelManyRefsOne',
        on_delete=models.CASCADE,
        related_name='two_1_x_b',
        null=True)
    
    one_x_x_b = models.ManyToManyField(
        'ModelManyRefsOne',
        related_name='two_x_x_b')
    
    one_1_1_b = models.OneToOneField(
        'ModelManyRefsOne',
        on_delete=models.CASCADE,
        related_name='two_1_1_b',
        null=True)


""" This module contains exceptions for
:py:class:`~save_to_db.core.utils.item_collection.ItemCollection` class.
"""

class ItemCollectionException(Exception):
    """ General exception for
    :py:class:`~save_to_db.core.utils.item_collection.ItemCollection`
    class.
    """


class CollectionDoesNotExist(ItemCollectionException):
    """ Raised when trying to get collection that does not exist using
    `collection_id`.
    """

class CollectionIdAlreadyInUse(ItemCollectionException):
    """ Raised when trying to create new item class collection with 
    `collection_id` value already taken.
    """


from save_to_db.exceptions import (ModelClsAlreadyRegistered,
                                   UnknownRelationFieldNames,
                                   UnknownModelDefaultKey,)


class MergePolicy(dict):
    """ Merge model policy defines how model conflicts should be resolved
    when using
    :py:meth:`~save_to_db.adapters.utils.adapter_base.AdapterBase.merge_models`
    method of
    :py:class:`~save_to_db.adapters.utils.adapter_base.AdapterBase` class.
    
    There are conflicts of two types:
    
        - Meging models have x-to-one relationship with other model class,
          but point to diffent model instances.

            Example:

                - Two merging model instances have many-to-one relationship
                  field `x_id` pointing to two different model X instances.
                  Merging the model requires them to point to the same instance
                  of X.

        - Meging models have one-to-many relationship with other model class X,
          but updating related model X instances to point to the merged instance
          breaks a unique constraint on model X.

            Example:

                - Two merging model instances of class A have one-to-many
                  relationship with model X and point to two different instances
                  of X. Model X has a unique constaint on field `a_id` pointing
                  to model A, thus making it impossible for two instances of X
                  to point to the same merged instance of A.
          
    :param db_adapter: An instance of
        :py:class:`~save_to_db.adapters.utils.adapter_base.AdapterBase`
        subclass.
    :param policy:
        A dictionary that looks like this:
        
            .. code-block:: Python
            
                merge_policy = {
                    OrmModelClass_1: {
                        x_to_1_field_1: resolve_x_to_1_field_1,
                        x_to_1_field_2: resolve_x_to_1_field_2,
                        1_to_many_field_1: resolve_1_to_many_field_1,
                        1_to_many_field_2: False,  # merge not allowed
                        ...
                    },
                    ...
                }
                
            **Where:**
        
                - *OrmModelClass_1* is a reference to an ORM model class.
                - *x_to_1_field_1* and *x_to_1_field_2* are x-to-one foreign
                  key field names from the class.
                - *1_to_many_field_1* and *1_to_many_field_2* are one-to-may
                  foreign key field names from the class.
                - *resolve_x_to_1_field_1* and *resolve_x_to_1_field_2* are
                  merge functions for the fields, function arguments:
                      
                    - *model* is model into which other model is merged.
                    - *merged_model* is a model that is in the process of being
                      merged.
                    - *child_model* is a model that *model* points to with
                      *forward_foreign_key*.
                      
                      *Here we refer to the model being
                      pointed to as "child", although in reality it can be
                      different.*
                    - *merged_child_model* is a model that *merged_model* points
                      to with *forward_foreign_key*.
                    - *forward_foreign_key* is a name of a parent field that
                      points to child.
                    - *reverse_foreign_key* is a name of a child field that
                      points parent.
                
                  *child_model* and *merged_child_model* are not the same, but
                  being pointed to by parent models from `models` argment using
                  the same key.
                  
                  The function must return one child model to be used.
                  
                  .. note::
                      *merged_child_model* is not deleted after the merging, for
                      x-to-one relationships it can be left unchanged.
                
                - *resolve_1_to_many_field_1* is a merge function for
                  the field. Conflicts in one-to-many fields may arise when
                  there is a unique constraint that is going to be broken if two
                  child models start to point to the same parent model.
                  The function accepts arguments the same arguments as
                  *resolve_x_to_1_field_1* and a list of unique
                  constaints (lists of field names) that are going to be broken.
                  
                  .. warning::
                      *merged_child_model* is not deleted after the merging.
                      Delete it yourself or make sure it does not brake the
                      unique constraints.
    
    :param defaults:
        Default model resolver. A dictionary that looks like this:
            
            .. code-block:: Python
            
                defaults = {
                    OrmModelClass_1: {
                        'one': resolve_x_to_1_field_1,
                        'many': resolve_1_to_many_field_1,
                    },
                    ...
                }
            
            **Where:**
                
                - *OrmModelClass_1*,
                - *resolve_x_to_1_field_1*,
                - and *'resolve_1_to_many_field_1* are same as for the
                  `policy` parameter.
            
        If, in the example above, some other ORM model class has a relationship
        with *OrmModelClass_1*, but does not define a resolver, default model
        resolver is used. 
    """
    
    #: Default key for x-to-one resolvers.
    REL_ONE = 'one'
    #: Default key for x-to-many resolvers.
    REL_MANY = 'many'
    
    def __init__(self, db_adapter, policy, defaults=None):
        defaults = defaults or {}
        self.db_adapter = db_adapter
        self.__validate(policy, defaults)
        self.defaults = defaults
        self.update(policy)
        self.__resolve_defaults()
        
        self.registered_models = set(policy.keys()).union(defaults.keys())
    
    
    def add_model_policy(self, model_cls, model_policy, model_defaults=None):
        """ Adds merge policy for a model.
        
        :param model_cls: An ORM model class to add.
        :param model_policy: Merge policy for the class.
            This has the same structure as `policy[model_cls]` (where
            `policy` is the argument passed to constructor).
        :param model_defaults: Default model resolvers for the class.
            This has the same structure as `defaults[model_cls]` (where
            `defaults` is the argument passed to constructor).
        """
        if model_cls in self.registered_models:
            raise ModelClsAlreadyRegistered(
                'Merge policy already registered for class {}'.format(model_cls)
            )
        
        model_defaults = model_defaults or {}
        self.__validate_model_policy(model_cls, model_policy)
        self.__validate_model_defaults(model_defaults)
        self[model_cls] = model_policy
        self.defaults[model_cls] = model_defaults
        self.__resolve_defaults()
        
        self.registered_models.add(model_cls)
    
    
    def __resolve_defaults(self):
        
        for model_cls in self.db_adapter.iter_all_models():
            for fkey, fmodel_cls, direction, _ in \
                    self.db_adapter.iter_relations(model_cls):
                
                if fmodel_cls not in self.defaults:
                    continue
                
                if direction.is_x_to_one():
                    multiplication_key = self.REL_ONE
                else:
                    multiplication_key = self.REL_MANY
                
                if multiplication_key not in self.defaults[fmodel_cls]:
                    continue
                
                if model_cls not in self:
                    self[model_cls] = {}
                
                # default must not overwrite
                if fkey not in self[model_cls]:
                    self[model_cls][fkey] = \
                        self.defaults[fmodel_cls][multiplication_key]
    
    
    def __validate(self, policy, defaults):
        for model_cls, model_policy in policy.items():
            self.__validate_model_policy(model_cls, model_policy)
        
        for model_cls, model_defaults in defaults.items():
            self.__validate_model_defaults(model_defaults)
    
    
    def __validate_model_policy(self, model_cls, model_policy):
        policy_fkeys = set(model_policy)
        model_fkeys = set(
            fkey for fkey, _, _, _ in self.db_adapter.iter_relations(model_cls)
        )
        unknown_fkeys = policy_fkeys - model_fkeys
        if unknown_fkeys:
            raise UnknownRelationFieldNames(model_cls, unknown_fkeys)
        
    
    def __validate_model_defaults(self, model_defaults):
        default_keys = set(model_defaults)
        known_keys = {self.REL_ONE, self.REL_MANY}
        unknown_fkeys = default_keys - known_keys
        if unknown_fkeys:
            raise UnknownModelDefaultKey(unknown_fkeys)


from save_to_db.core.item import Item
from save_to_db.core.merge_policy import MergePolicy
from save_to_db.utils.test_base import TestBase

from save_to_db.exceptions import (ModelClsAlreadyRegistered,
                                   UnknownRelationFieldNames,
                                   UnknownModelDefaultKey,)


class TestPolicySetup(TestBase):
    
    def setUp(self):
        super().setUp()
        
        class ItemMergeOne(Item):
            model_cls = self.ModelMergeOne
        self.ItemMergeOne = ItemMergeOne
        
        class ItemMergeTwo(Item):
            model_cls = self.ModelMergeTwo
        self.ItemMergeTwo = ItemMergeTwo
        
        class ItemMergeThree(Item):
            model_cls = self.ModelMergeThree
        self.ItemMergeThree = ItemMergeThree
        
        
        class ItemMergeOneX(Item):
            model_cls = self.ModelMergeOneX
        self.ItemMergeOneX = ItemMergeOneX
        
        class ItemMergeTwoX(Item):
            model_cls = self.ModelMergeTwoX
        self.ItemMergeTwoX = ItemMergeTwoX
        
        class ItemMergeThreeX(Item):
            model_cls = self.ModelMergeThreeX
        self.ItemMergeThreeX = ItemMergeThreeX
    
    
   
    def test_creation_and_addition(self):
        # NOTE: In this test we are using string values as dunny functions

        #--- creation test ---
        merge_policy_one = MergePolicy(
            self.db_adapter,
            policy={
                self.ModelMergeOne: {
                    'two_1_1': 'resolve_two_one',
                },
                self.ModelMergeThree: {
                    'two_1_x': 'resolve_two_many',
                },
            },
            defaults={
                self.ModelMergeOne: {
                    MergePolicy.REL_ONE: 'resolve_one_one_default',
                    MergePolicy.REL_MANY: 'resolve_one_many_default',
                },
            })
        expected_one = {
            self.ModelMergeOne: {
                'two_1_1': 'resolve_two_one',
            },
            # defaults
            self.ModelMergeTwo: {
                'one_1_1': 'resolve_one_one_default',
                'one_1_x': 'resolve_one_many_default',
            },
            self.ModelMergeThree: {
                'two_1_x': 'resolve_two_many',
            }
        }
        self.assertEqual(merge_policy_one, expected_one)
         
        merge_policy_two = MergePolicy(
            self.db_adapter,
            policy={
                self.ModelMergeOne: {
                    'two_1_1': 'resolve_two_one',
                },
            },
            defaults={
                self.ModelMergeTwo: {
                    MergePolicy.REL_ONE: 'resolve_two_one_default',
                },
            })
        expected_two = {
            self.ModelMergeOne: {
                'two_1_1': 'resolve_two_one',  # default must not overwrite
                'two_x_1': 'resolve_two_one_default',
            },
            self.ModelMergeThree: {
                'two_1_1': 'resolve_two_one_default',
            }
        }
        self.assertEqual(merge_policy_two, expected_two)
        
        #--- addition test ---
        
        # overwriting model two
        merge_policy_one.add_model_policy(
            self.ModelMergeTwo,
            model_policy={
                'one_1_1': 'resolve_one_one',
            },
            model_defaults={
                MergePolicy.REL_ONE: 'resolve_two_one_default',
            }
        )
        expected_one_new = {
            self.ModelMergeOne: {
                'two_1_1': 'resolve_two_one',
                'two_x_1': 'resolve_two_one_default',  # injected new
            },
            # defaults
            self.ModelMergeTwo: {
                'one_1_1': 'resolve_one_one',  # overwritten
                'one_1_x': 'resolve_one_many_default',  # injected old
            },
            self.ModelMergeThree: {
                'two_1_1': 'resolve_two_one_default',  # injected new
                'two_1_x': 'resolve_two_many',
            }
        }
        self.assertEqual(merge_policy_one, expected_one_new)
    
        # adding model two
        merge_policy_two.add_model_policy(
            self.ModelMergeThree,
            model_policy={
                'two_1_1': 'resolve_two_one',
            },
            model_defaults={
                MergePolicy.REL_ONE: 'resolve_three_one_default',
            }
        )
        expect_two_new = {
            self.ModelMergeOne: {
                'two_1_1': 'resolve_two_one',
                'two_x_1': 'resolve_two_one_default',
            },
            self.ModelMergeTwo: {
                'three_1_1': 'resolve_three_one_default',  # added
                'three_x_1': 'resolve_three_one_default',  # added
            },
            self.ModelMergeThree: {
                'two_1_1': 'resolve_two_one',  # overwritten
            }
        }
        self.assertEqual(merge_policy_two, expect_two_new)
    
    
    def test_exceptions(self):
        
        def create_policy():
            return MergePolicy(
                self.db_adapter,
                policy={
                    self.ModelMergeOne: {
                        'two_1_1': 'resolve_two_one',
                    },
                },
                defaults={
                    self.ModelMergeTwo: {
                        MergePolicy.REL_ONE: 'resolve_two_one_default',
                    },
                })
        
        #--- adding already resitered classes ---
        merge_policy = create_policy()
        
        # in policy
        with self.assertRaises(ModelClsAlreadyRegistered):
            merge_policy.add_model_policy(self.ModelMergeOne, model_policy={})
        
        # in defaults
        with self.assertRaises(ModelClsAlreadyRegistered):
            merge_policy.add_model_policy(self.ModelMergeTwo, model_policy={})
        
        #--- using wrong relation field names ---
        with self.assertRaises(UnknownRelationFieldNames):
            merge_policy = MergePolicy(
                self.db_adapter,
                policy={
                    self.ModelMergeOne: {
                        'unknown_field': 'resolve_unknown_field',
                    },
                })
        with self.assertRaises(UnknownRelationFieldNames):
            merge_policy.add_model_policy(
                self.ModelMergeThree,
                model_policy={
                    'unknown_field': 'resolve_unknown_field',
                })
            
        with self.assertRaises(UnknownRelationFieldNames):
            merge_policy = MergePolicy(
                self.db_adapter,
                policy={
                    self.ModelMergeOne: {
                        'f_integer': 'resolve_f_integer',  # not a relation
                    },
                })
        with self.assertRaises(UnknownRelationFieldNames):
            merge_policy.add_model_policy(
                self.ModelMergeThree,
                model_policy={
                    'f_integer': 'resolve_f_integer',  # not a relation
                })
            
        #--- wrong key for defaults
        with self.assertRaises(UnknownModelDefaultKey):
            merge_policy = MergePolicy(
                self.db_adapter,
                policy={},
                defaults={
                    self.ModelMergeTwo: {
                        'wrong_key': 'resolve_model_two',
                    },
                })
        with self.assertRaises(UnknownModelDefaultKey):
            merge_policy.add_model_policy(
                self.ModelMergeThree,
                model_policy={},
                model_defaults={
                    'wrong_key': 'resolve_model_two',
                })
        
        
        # merge policy must be the same as at the beginning
        clean_policy = create_policy()
        self.assertEqual(merge_policy, clean_policy)
        self.assertEqual(merge_policy.defaults, clean_policy.defaults)
        self.assertEqual(merge_policy.registered_models,
                         clean_policy.registered_models)


from save_to_db.core import signals
from save_to_db.exceptions.item_persist import \
    CannotClearRequiredFieldInRelation
from .model_updater import ModelUpdater
from .match_items_to_models import match_items_to_models


def db_persist(db_adapter, item):
    top_item = item
    result_items = item.as_list()
    
    item_structure = item.process()
    
    # emitting `before_db_persist` signal
    signals.before_db_persist.emit(top_item, item_structure)
    
    # creating model updater
    model_updater = ModelUpdater(item_structure, db_adapter)
    process_keeper = model_updater.process_keeper
    
    # --- getting and updating initial models ---
    getter_process_keeper = process_keeper
    while not getter_process_keeper.is_empty():
        all_items, all_models = [], []
        for item_cls in getter_process_keeper:
            if item_cls.fast_insert or not getter_process_keeper[item_cls]:
                continue
            
            items_and_fkeys = []
            new_items, new_models = [], []
            
            for item_track in getter_process_keeper[item_cls]:
                items_and_fkeys.append([item_track.item,
                                        item_track.fkeys])
            
            batch_size = item_cls.batch_size \
                if item_cls.batch_size is not None \
                else db_adapter.BATCH_SIZE
            got_models = []
            for batch_start in range(0, len(items_and_fkeys), batch_size):
                got_models.extend(
                    db_adapter.get(
                        items_and_fkeys[batch_start:batch_start+batch_size]
                    )
                )
            if not got_models:
                continue
            got_items, got_models = match_items_to_models(db_adapter,
                                                          items_and_fkeys,
                                                          got_models)
            all_items.extend(got_items)
            all_models.extend(got_models)
        
        getter_process_keeper = model_updater.update_relationships(
            all_items, all_models)[1]
    
    for item_cls in process_keeper:
        for item_track in process_keeper[item_cls]:
            item = item_track.item
            models = item_track.models
            for model in models:
                item.before_model_update(model)  # hook
                
                # cleaning x-to-many relations if needed
                for fkey in item_cls.relations:
                    relation = item_cls.relations[fkey]
                    if not relation['relation_type'].is_x_to_many() or \
                            not relation['replace_x_to_many'] or \
                            fkey not in item:
                        continue
                    if not model_updater.can_set_model_field(
                            item, model, fkey, None):
                        continue
                    
                    if 'reverse_key' in relation:
                        reverse_key = relation['reverse_key']
                        other_item_cls = relation['item_cls']
                        
                        for creator_group in other_item_cls.creators:
                            if reverse_key in creator_group:
                                raise CannotClearRequiredFieldInRelation(
                                    'Cannot clear required field.')
                    
                    model_updater.clear_related_models(item, model, fkey)
                
                model_updater.update_model_fields(item_track, model)
                
    # --- creation loop ---
    create_process_keeper = process_keeper
    while not create_process_keeper.is_empty():
        new_items, new_models = [], []
        for item_cls in create_process_keeper:
            for item_track in create_process_keeper[item_cls]:
                if item_track.models:
                    continue
                item = item_track.item
                if item not in new_items and \
                        model_updater.can_create_model(item, item_track.fkeys):
                    new_items.append(item)
                    model = db_adapter.create_blank_model(item.model_cls)
                    model_updater.add_created_model(model)
                    new_models.append([model])
                    item.before_model_update(model)  # hook
                    model_updater.update_model_fields(item_track, model)

        create_process_keeper = model_updater.update_relationships(
            new_items, new_models)[0]
    
    # emitting `item_dropped` signal
    for item_cls in process_keeper:
        for item_track in process_keeper[item_cls]:
            if item_track.models:
                continue
            item = item_track.item
            if item.get_only_mode:
                reason = \
                    signals.item_dropped.reason_cannot_create_get_only_mode
            elif item.update_only_mode:
                reason = \
                    signals.item_dropped.reason_cannot_create_update_only_mode
            else:
                reason = \
                    signals.item_dropped.reason_cannot_create_not_enough_data
            
            signals.item_dropped.emit(item, reason)
    
    # saving models
    for item_cls in process_keeper:
        for item_track in process_keeper[item_cls]:
            item = item_track.item
            for model in item_track.models:
                db_adapter.save_model(model)
                
                if item_cls.metadata['model_deleter'] is not None:
                    item_cls.metadata['model_deleter'].collect_model(model)
                
    # executing model unrefs
    for item_cls in process_keeper:
        for item_track in process_keeper[item_cls]:
            item = item_track.item
            for fkey, model_deleter in item.metadata['model_unrefs'].items():
                for parent_model in item_track.models:
                    model_deleter.execute_unref(parent_model, fkey, db_adapter)
    
    # `after_model_save` hook
    for item_cls in process_keeper:
        for item_track in process_keeper[item_cls]:
            item = item_track.item
            for model in item_track.models:
                item.after_model_save(model)
    
    # creating result
    items, models = [], []
    for item in result_items:
        # looking for models
        for item_track in process_keeper[type(item)]:
            if item is item_track.item:
                if item_track.models:
                    items.append(item_track.item)
                    models.append(item_track.models)
                break
    
    # emitting `after_db_persist` signal
    signals.after_db_persist.emit(top_item, items, models)
    
    # executing model deleters
    for item_cls in item_structure:
        if item_cls.deleter_execute_on_persist and \
                item_cls.metadata['model_deleter']:
            item_cls.metadata['model_deleter'].execute_delete(db_adapter)
    
    return items, models




import sqlalchemy as sa
from sqlalchemy.orm import backref, relationship
from sqlalchemy.schema import ForeignKeyConstraint, UniqueConstraint

from ..config import Base



class __CommonColumns(object):
    f_binary = sa.Column(sa.LargeBinary, nullable=True)
    f_boolean = sa.Column(sa.Boolean, nullable=True)
    f_string = sa.Column(sa.String, nullable=True)
    f_text = sa.Column(sa.Text, nullable=True)
    f_integer = sa.Column(sa.Integer, nullable=True)
    f_float = sa.Column(sa.Float, nullable=True)
    f_date = sa.Column(sa.Date, nullable=True)
    f_time = sa.Column(sa.Time, nullable=True)
    f_datetime = sa.Column(sa.DateTime, nullable=True)


class ModelConstraintsOne(__CommonColumns, Base):
    __tablename__ = 'model_constraints_one'
    
    id = sa.Column(sa.Integer, primary_key=True)
    
    f_text = sa.Column(sa.Text, nullable=False)
    f_integer = sa.Column(sa.Integer, unique=True)
    f_string = sa.Column(sa.String, nullable=False, unique=True)
    
    two_1_x = relationship(
        'ModelConstraintsTwo',
        back_populates='one_x_1',
        foreign_keys='[ModelConstraintsTwo.one_x_1_id]')
    
    three_1_x = relationship(
        'ModelConstraintsThree',
        back_populates='one_x_1',
        foreign_keys='[ModelConstraintsThree.one_x_1_id]')
    
    five_1_1 = relationship(
        'ModelConstraintsFive',
        back_populates='one_1_1',
        foreign_keys='[ModelConstraintsFive.one_1_1_id]',
        uselist=False)


class ModelConstraintsTwo(__CommonColumns, Base):
    __tablename__ = 'model_constraints_two'
    
    id = sa.Column(sa.Integer, primary_key=True)
    
    one_x_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_constraints_one.id', ondelete='CASCADE'))
    one_x_1 = relationship(
        ModelConstraintsOne,
        back_populates='two_1_x',
        foreign_keys=[one_x_1_id])
    
    three_x_x = relationship(
        'ModelConstraintsThree',
        secondary='model_constraints_two_to_three',
        back_populates='two_x_x')

    four_primary_one = sa.Column(sa.Integer, nullable=False)
    four_primary_two = sa.Column(sa.Integer, nullable=False)
    four_x_1 = relationship(
        'ModelConstraintsFour',
        back_populates='two_1_x',
        foreign_keys=[four_primary_one, four_primary_two])
    
    __table_args__ = (
        ForeignKeyConstraint([four_primary_one, four_primary_two],
                             ['model_constraints_four.primary_one',
                              'model_constraints_four.primary_two'],
                             name='model_four_fkey'),
    )
    

class ModelConstraintsTwoToThree(Base):
    __tablename__ = 'model_constraints_two_to_three'
    
    two_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(ModelConstraintsTwo.id, ondelete='CASCADE'),
        primary_key=True)
    three_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_constraints_three.id', ondelete='CASCADE'),
        primary_key=True)
    

class ModelConstraintsThree(__CommonColumns, Base):
    __tablename__ = 'model_constraints_three'
    
    id = sa.Column(sa.Integer, primary_key=True)
    
    one_x_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_constraints_one.id', ondelete='CASCADE'),
        nullable=False)
    one_x_1 = relationship(
        ModelConstraintsOne,
        back_populates='three_1_x',
        foreign_keys=[one_x_1_id])
    
    two_x_x = relationship(
        ModelConstraintsTwo,
        secondary='model_constraints_two_to_three',
        back_populates='three_x_x')
    
    four_primary_one = sa.Column(sa.Integer, nullable=True)
    four_primary_two = sa.Column(sa.Integer, nullable=True)
    four_x_1 = relationship(
        'ModelConstraintsFour',
        back_populates='three_1_x',
        foreign_keys=[four_primary_one, four_primary_two])
    
    __table_args__ = (
        ForeignKeyConstraint([four_primary_one, four_primary_two],
                             ['model_constraints_four.primary_one',
                              'model_constraints_four.primary_two'],
                             name='model_four_fkey'),
    )


class ModelConstraintsFour(__CommonColumns, Base):
    __tablename__ = 'model_constraints_four'
    
    primary_one = sa.Column(sa.Integer, primary_key=True)
    primary_two = sa.Column(sa.Integer, primary_key=True)
    
    two_1_x = relationship(
        ModelConstraintsTwo,
        back_populates='four_x_1',
        foreign_keys=[ModelConstraintsTwo.four_primary_one,
                      ModelConstraintsTwo.four_primary_two,])
    
    three_1_x = relationship(
        ModelConstraintsThree,
        back_populates='four_x_1',
        foreign_keys=[ModelConstraintsThree.four_primary_one,
                      ModelConstraintsThree.four_primary_two,])
    
    five_1_1 = relationship(
        'ModelConstraintsFive',
        back_populates='four_1_1',
        foreign_keys='[ModelConstraintsFive.four_primary_one,'
                      'ModelConstraintsFive.four_primary_two,]',
        uselist=False)
    
    __table_args__ = (
        UniqueConstraint('f_integer', 'f_string', name='unique_1'),
    )
    

class ModelConstraintsFive(__CommonColumns, Base):
    __tablename__ = 'model_constraints_five'
    
    id = sa.Column(sa.Integer, primary_key=True)
    
    one_1_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_constraints_one.id', ondelete='CASCADE'),
        unique=True)
    one_1_1 = relationship(
        ModelConstraintsOne,
        back_populates='five_1_1',
        foreign_keys=[one_1_1_id])
    
    four_primary_one = sa.Column(sa.Integer)
    four_primary_two = sa.Column(sa.Integer)
    four_1_1 = relationship(
        ModelConstraintsFour,
        back_populates='five_1_1',
        foreign_keys=[four_primary_one, four_primary_two])
    
    six_1_1 = relationship(
        'ModelConstraintsSix',
        back_populates='five_1_1',
        foreign_keys='[ModelConstraintsSix.five_1_1_id,]',
        uselist=False)
    
    __table_args__ = (
        UniqueConstraint('four_primary_one', 'four_primary_two',
                         name='unique_1'),
        ForeignKeyConstraint([four_primary_one, four_primary_two],
                             ['model_constraints_four.primary_one',
                              'model_constraints_four.primary_two'],
                             name='model_four_fkey'),
    )

class ModelConstraintsSix(__CommonColumns, Base):
    __tablename__ = 'model_constraints_six'
    
    id = sa.Column(sa.Integer, primary_key=True)
    
    five_1_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_constraints_five.id', ondelete='CASCADE'),
        nullable=False)
    five_1_1 = relationship(
        ModelConstraintsFive,
        back_populates='six_1_1',
        foreign_keys=[five_1_1_id])
    
    __table_args__ = (
        UniqueConstraint('five_1_1_id', 'f_integer', name='unique_1'),
    )
    
    
class ModelConstraintsSelf(__CommonColumns, Base):
    __tablename__ = 'model_constraints_self'
    
    code = sa.Column(sa.String, primary_key=True)
    
    # many-to-one, not_null
    parent_x_1_code = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_constraints_self.code', ondelete='CASCADE'),
        nullable=False)
    child_1_x = relationship(
        'ModelConstraintsSelf',
        backref=backref('parent_x_1', remote_side=[code]),
        foreign_keys=[parent_x_1_code])
    
    # self one-to-one, unique
    first_parent_1_1_code = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_constraints_self.code', ondelete='CASCADE'),
        unique=True)
    first_child_1_1 = relationship(
        'ModelConstraintsSelf',
        backref=backref('first_parent_1_1', remote_side=[code]),
        foreign_keys=[first_parent_1_1_code],
        uselist=False)
    
    # self one-to-one, unique, not_null
    second_parent_1_1_code = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_constraints_self.code', ondelete='CASCADE'),
        unique=True,
        nullable=False)
    second_child_1_1 = relationship(
        'ModelConstraintsSelf',
        backref=backref('second_parent_1_1', remote_side=[code]),
        foreign_keys=[second_parent_1_1_code],
        uselist=False)
    
    # self many-to-many
    parent_x_x = relationship(
        'ModelConstraintsSelf',
        secondary='model_constraints_self_to_self',
        foreign_keys='[ModelConstraintsSelfToSelf.parent_code]',
        back_populates='child_x_x')
    child_x_x = relationship(
        'ModelConstraintsSelf',
        secondary='model_constraints_self_to_self',
        foreign_keys='[ModelConstraintsSelfToSelf.child_code]',
        back_populates='parent_x_x')


class ModelConstraintsSelfToSelf(__CommonColumns, Base):
    __tablename__ = 'model_constraints_self_to_self'
    
    parent_code = sa.Column(
        sa.Integer,
        sa.ForeignKey(ModelConstraintsSelf.code, ondelete='CASCADE'),
        primary_key=True)
    parent = relationship(
        ModelConstraintsSelf,
        foreign_keys=[parent_code])
    
    child_code = sa.Column(
        sa.Integer,
        sa.ForeignKey(ModelConstraintsSelf.code, ondelete='CASCADE'),
        primary_key=True)
    child = relationship(
        ModelConstraintsSelf,
        foreign_keys=[child_code])


from save_to_db.core.item import Item
from save_to_db.core.utils.mapper import mapper 
from save_to_db.utils.test_base import TestBase


class TestMapper(TestBase):
    
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        
        class ItemGeneralOne(Item):
            model_cls = cls.ModelGeneralOne
                
        class ItemGeneralTwo(Item):
            model_cls = cls.ModelGeneralTwo
        
        cls.ItemGeneralOne = ItemGeneralOne
        cls.ItemGeneralTwo = ItemGeneralTwo
    
    
    def test_item_added_on_creation(self):
        single_item = self.ItemGeneralOne()
        self.assertIn(single_item, mapper)
        
        bulk_item = self.ItemGeneralOne.Bulk()
        self.assertIn(bulk_item, mapper)
    
    
    def test_reverse_key_one_to_one(self):
        #--- setting ---
        # direct set ...
        item_one = self.ItemGeneralOne()
        first_item_two = self.ItemGeneralTwo()
        item_one['two_1_1'] = first_item_two
         
        self.assertIn('one_1_1', first_item_two)
        self.assertIs(first_item_two['one_1_1'], item_one)
        
        # ... direct overwrite
        second_item_two = self.ItemGeneralTwo()
        item_one['two_1_1'] = second_item_two
        
        self.assertIn('one_1_1', second_item_two)
        self.assertIs(second_item_two['one_1_1'], item_one)
        
        self.assertNotIn('one_1_1', first_item_two)
        
        # set on creation
        item_two = self.ItemGeneralTwo()
        item_one = self.ItemGeneralOne(two_1_1=item_two)
        
        self.assertIn('one_1_1', item_two)
        self.assertIs(item_two['one_1_1'], item_one)
        
        # set on getting
        item_one = self.ItemGeneralOne()
        item_two = item_one['two_1_1']
        
        self.assertIn('one_1_1', item_two)
        self.assertIs(item_two['one_1_1'], item_one)

        # set by call ...
        first_item_two = self.ItemGeneralTwo()
        item_one = self.ItemGeneralOne()
        item_one(two_1_1=first_item_two)
        
        self.assertIn('one_1_1', first_item_two)
        self.assertIs(first_item_two['one_1_1'], item_one)
        
        # ... overwrite by call
        second_item_two = self.ItemGeneralTwo()
        item_one(two_1_1=second_item_two)
        
        self.assertIn('one_1_1', second_item_two)
        self.assertIs(second_item_two['one_1_1'], item_one)
        
        self.assertNotIn('one_1_1', first_item_two)
        
        #--- overwrite by `None` ---
        item_one = self.ItemGeneralOne()
        item_two = self.ItemGeneralTwo()
        item_one['two_1_1'] = item_two
         
        item_one['two_1_1'] = None

        self.assertNotIn('one_1_1', item_two)
        
        #--- deleting item value ---
        item_one = self.ItemGeneralOne()
        item_two = self.ItemGeneralTwo()
        item_one['two_1_1'] = item_two
         
        del item_one['two_1_1']
        
        self.assertNotIn('two_1_1', item_one)
        self.assertNotIn('one_1_1', item_two)
                
        #--- deleting item itsef ---
        item_one = self.ItemGeneralOne()
        item_two = self.ItemGeneralTwo(f_integer=1)
        item_one['two_1_1'] = item_two
        
        self.assertIn('one_1_1', item_two)
        item_two.delete()
        
        self.assertNotIn('one_1_1', item_two)
        self.assertNotIn('f_integer', item_two)
         
        self.assertNotIn('two_1_1', item_one)

    
    def test_reverse_key_one_to_many(self):
        #--- setting ---
         
        # set and add ...
        item_one = self.ItemGeneralOne()
        first_bulk_two = self.ItemGeneralTwo.Bulk()
        item_one['two_1_x'] = first_bulk_two
          
        first_item_two = self.ItemGeneralTwo()
        first_bulk_two.add(first_item_two)
         
        self.assertIn('one_x_1', first_item_two)
        self.assertIs(first_item_two['one_x_1'], item_one)
         
        # ... direct overwrite
        second_bulk_two = self.ItemGeneralTwo.Bulk()
        second_item_two = self.ItemGeneralTwo()
        second_bulk_two.add(second_item_two)
        item_one['two_1_x'] = second_bulk_two
         
        self.assertNotIn('one_x_1', first_item_two)  # not in first bulk
        self.assertIn('one_x_1', second_item_two)    # in second bulk
        self.assertIs(second_item_two['one_x_1'], item_one)
        
        # gen
        item_one = self.ItemGeneralOne()
        bulk_two = self.ItemGeneralTwo.Bulk()
        item_one['two_1_x'] = bulk_two
          
        item_two = bulk_two.gen()
         
        self.assertIn('one_x_1', item_two)
        self.assertIs(item_two['one_x_1'], item_one)
         
        #--- removing ---
        item_one = self.ItemGeneralOne()
        bulk_two = self.ItemGeneralTwo.Bulk()
        item_one['two_1_x'] = bulk_two
          
        first_item_two = bulk_two.gen()
        second_item_two = bulk_two.gen()
         
        self.assertIs(first_item_two['one_x_1'], item_one)
        self.assertIs(second_item_two['one_x_1'], item_one)
         
        bulk_two.remove(first_item_two)
         
        self.assertNotIn('one_x_1', first_item_two)
        self.assertIs(second_item_two['one_x_1'], item_one)
         
        #--- deleting ---
         
        # bulk reference
        item_one = self.ItemGeneralOne()
        bulk_two = self.ItemGeneralTwo.Bulk()
        item_one['two_1_x'] = bulk_two
          
        first_item_two = bulk_two.gen()
        second_item_two = bulk_two.gen()
         
        self.assertIn(first_item_two, bulk_two)
        self.assertIn(second_item_two, bulk_two)
         
        del item_one['two_1_x']
         
        self.assertIn(first_item_two, bulk_two)
        self.assertIn(second_item_two, bulk_two)
         
        self.assertNotIn('one_x_1', first_item_two)
        self.assertNotIn('one_x_1', second_item_two)
        
        # parent item
        item_one = self.ItemGeneralOne()
        bulk_two = self.ItemGeneralTwo.Bulk()
        item_one['two_1_x'] = bulk_two
          
        first_item_two = bulk_two.gen()
        second_item_two = bulk_two.gen()
        
        item_one.delete()
        
        self.assertNotIn('two_1_x', item_one)
        
        self.assertIn(first_item_two, bulk_two)
        self.assertIn(second_item_two, bulk_two)
        
        self.assertNotIn('one_x_1', first_item_two)
        self.assertNotIn('one_x_1', second_item_two)
        
        
        # bulk itself
        item_one = self.ItemGeneralOne()
        bulk_two = self.ItemGeneralTwo.Bulk()
        item_one['two_1_x'] = bulk_two
          
        first_item_two = bulk_two.gen()
        second_item_two = bulk_two.gen()
         
        self.assertIn(first_item_two, bulk_two)
        self.assertIn(second_item_two, bulk_two)
         
        self.assertIs(first_item_two['one_x_1'], item_one)
        self.assertIs(second_item_two['one_x_1'], item_one)
         
        bulk_two.delete()

        self.assertNotIn('two_1_x', item_one)
        self.assertEqual(len(bulk_two), 0)
          
        self.assertNotIn('one_x_1', first_item_two)
        self.assertNotIn('one_x_1', second_item_two)
        
        # item in bulk
        item_one = self.ItemGeneralOne()
        bulk_two = self.ItemGeneralTwo.Bulk()
        item_one['two_1_x'] = bulk_two
          
        first_item_two = bulk_two.gen()
        second_item_two = bulk_two.gen()
         
        self.assertIn(first_item_two, bulk_two)
        self.assertIn(second_item_two, bulk_two)
         
        first_item_two.delete()
         
        self.assertNotIn(first_item_two, bulk_two)
        self.assertIn(second_item_two, bulk_two)
    
    
    def test_reverse_key_many_to_many(self):
        #--- setting ---
        
        # direct set
        item_one = self.ItemGeneralOne()
        bulk_two = self.ItemGeneralTwo.Bulk()
        item_one['two_x_x'] = bulk_two
         
        # gen item on that side
        genned_item_two = bulk_two.gen()
        self.assertIn('one_x_x', genned_item_two)
        self.assertIn(item_one, genned_item_two['one_x_x'])
        self.assertIn(genned_item_two, item_one['two_x_x'])
         
        # gen item on this side
        genned_item_one = genned_item_two['one_x_x'].gen()
         
        self.assertIn('two_x_x', genned_item_one)
        self.assertIn(genned_item_two, genned_item_one['two_x_x'])
        self.assertIn(genned_item_one, genned_item_two['one_x_x'])
         
        # --- overwriting with non-empty bulk-items ---
        bulk_one = self.ItemGeneralOne.Bulk()
          
        bulk_two = self.ItemGeneralTwo.Bulk()
        item_two = bulk_two.gen(f_integer=200)
          
        item_one = bulk_one.gen(f_integer=100)
        item_one['two_x_x'] = bulk_two
        
        # reverse was set
        self.assertIn(item_one, item_two['one_x_x'])
        
        second_bulk_two = self.ItemGeneralTwo.Bulk()
        item_one['two_x_x'] = second_bulk_two
        
        # reverse cleared
        self.assertNotIn(item_one, item_two['one_x_x'])
         
        #--- separate many-to-many ---
        # [REF: 1]
        # Items:
        #    one: 11, 12
        #    two: 21, 22
        #
        #    11 has:         21, 22
        #    21 has: 11, 12
        #    (reverse)
        #    12 has:             21
        #    22 has: 11
        item_11 = self.ItemGeneralOne(f_integer=11)
        item_12 = self.ItemGeneralOne(f_integer=12)
        item_21 = self.ItemGeneralTwo(f_integer=21)
        item_22 = self.ItemGeneralTwo(f_integer=22)
         
        item_11['two_x_x'].add(item_21, item_22)
        item_21['one_x_x'].add(item_11, item_12)
         
        self.assertEqual(len(item_11['two_x_x']), 2)
        self.assertIn(item_21, item_11['two_x_x'])
        self.assertIn(item_22, item_11['two_x_x'])
         
        self.assertEqual(len(item_21['one_x_x']), 2)
        self.assertIn(item_11, item_21['one_x_x'])
        self.assertIn(item_12, item_21['one_x_x'])
         
        # (reverse)
        self.assertEqual(len(item_12['two_x_x']), 1)
        self.assertIn(item_21, item_12['two_x_x'])
         
        self.assertEqual(len(item_22['one_x_x']), 1)
        self.assertIn(item_11, item_22['one_x_x'])
         
        #--- removing items ---
        # [REF: 2]
        #    removing from 11: 21
        #
        #    (removed items in brackets)
        #    11 has:            [21], 22
        #    21 has: [11],  12
        #    (reverse)
        #    12 has:                  21
        #    22 has:  11
        item_11['two_x_x'].remove(item_21)
         
        self.assertEqual(len(item_11['two_x_x']), 1)
        self.assertIn(item_22, item_11['two_x_x'])
         
        self.assertEqual(len(item_21['one_x_x']), 1)
        self.assertIn(item_12, item_21['one_x_x'])
          
        self.assertEqual(len(item_12['two_x_x']), 1)
        self.assertIn(item_21, item_12['two_x_x'])
          
        self.assertEqual(len(item_22['one_x_x']), 1)
        self.assertIn(item_11, item_22['one_x_x'])
        
        #--- deleting ---
        
        # creates structure as on [REF: 1]
        item_11, item_12, item_21, item_22 = None, None, None, None
        def make_structure():
            nonlocal item_11, item_12, item_21, item_22
            item_11 = self.ItemGeneralOne(f_integer=11)
            item_12 = self.ItemGeneralOne(f_integer=12)
            item_21 = self.ItemGeneralTwo(f_integer=21)
            item_22 = self.ItemGeneralTwo(f_integer=22)
        
            item_11['two_x_x'].add(item_21, item_22)
            item_21['one_x_x'].add(item_11, item_12)
        
        # deleting bulk in 11: (removed items in brackets)
        # 
        #    11 has:            [21], [22] - deleted (reset)
        #    21 has: [11],  12
        #    (reverse)
        #    12 has:             21
        #    22 has: [11]

        # deleting bulk item
        make_structure()
        item_11['two_x_x'].delete()
         
        self.assertNotIn('two_x_x', item_11)
         
        self.assertEqual(len(item_21['one_x_x']), 1)
        self.assertIn(item_12, item_21['one_x_x'])
 
        self.assertEqual(len(item_12['two_x_x']), 1)
        self.assertIn(item_21, item_12['two_x_x'])
           
        self.assertIn('one_x_x', item_22)
        self.assertEqual(len(item_22['one_x_x']), 0)
        
        # deleting reference to bulk
        make_structure()
        del item_11['two_x_x']
        
        self.assertNotIn('two_x_x', item_11)
         
        self.assertEqual(len(item_21['one_x_x']), 1)
        self.assertIn(item_12, item_21['one_x_x'])
 
        self.assertEqual(len(item_12['two_x_x']), 1)
        self.assertIn(item_21, item_12['two_x_x'])
           
        self.assertIn('one_x_x', item_22)
        self.assertEqual(len(item_22['one_x_x']), 0)
        
        
        # deleting single item
        make_structure()
        item_11.delete()
        
        # all many-to-many must stay in play, but item deleted from them
        self.assertNotIn('two_x_x', item_11)  # reset
         
        self.assertEqual(len(item_21['one_x_x']), 1)
        self.assertIn(item_12, item_21['one_x_x'])
         
        # (reverse)
        self.assertEqual(len(item_12['two_x_x']), 1)
        self.assertIn(item_21, item_12['two_x_x'])
         
        self.assertIn('one_x_x', item_22)
        self.assertEqual(len(item_22['one_x_x']), 0)
        
    
    def test_reverse_key_many_to_one(self):
        # --- set direct ---
        
        # item two without reverse
        item_one = self.ItemGeneralOne()
        item_two = self.ItemGeneralTwo()
        item_one['two_x_1'] = item_two
        self.assertEqual(len(item_two['one_1_x']), 1)
        self.assertIn(item_one, item_two['one_1_x'])
         
        # overwriting
        second_item_two = self.ItemGeneralTwo()
        item_one['two_x_1'] = second_item_two
        self.assertEqual(len(second_item_two['one_1_x']), 1)
        self.assertIn(item_one, second_item_two['one_1_x'])
        self.assertEqual(len(item_two['one_1_x']), 0)
        self.assertNotIn(item_one, item_two['one_1_x'])

        # item two with reverse
        item_one = self.ItemGeneralOne()
        item_two = self.ItemGeneralTwo(one_1_x=[])
        item_one['two_x_1'] = item_two
        self.assertEqual(len(item_two['one_1_x']), 1)
        self.assertIn(item_one, item_two['one_1_x'])

        # --- deleting ---
        
        # deleting item that item
        item_one = self.ItemGeneralOne()
        item_two = self.ItemGeneralTwo(one_1_x=[])
        item_one['two_x_1'] = item_two
        
        item_two.delete()
        
        self.assertNotIn('one_1_x', item_two)
        self.assertNotIn('two_x_1', item_one)
        
        # deleting this that item
        item_one = self.ItemGeneralOne()
        item_two = self.ItemGeneralTwo()
        item_one['two_x_1'] = item_two
        
        item_one.delete()
        
        self.assertIn('one_1_x', item_two)
        self.assertNotIn(item_one, item_two['one_1_x'])
        
        # deleting reference        
        item_one = self.ItemGeneralOne()
        item_two = self.ItemGeneralTwo()
        item_one['two_x_1'] = item_two
        
        del item_one['two_x_1']
        
        self.assertIn('one_1_x', item_two)
        self.assertNotIn(item_one, item_two['one_1_x'])
        
        # deleting bulk
        item_one = self.ItemGeneralOne()
        item_two = self.ItemGeneralTwo()
        item_one['two_x_1'] = item_two
        
        item_two['one_1_x'].delete()
        
        self.assertEqual(len(item_two['one_1_x']), 0)
        self.assertNotIn('two_x_1', item_one)
    
    
    def test_item_delete(self):
        # --- single item ---
        two_1_x = self.ItemGeneralTwo.Bulk()
        two_1_x.gen()
        two_1_x.gen()
        
        two_x_x = self.ItemGeneralTwo.Bulk()
        two_x_x.gen()
        two_x_x.gen()
        
        item_one_fields = {
            'f_integer': 10,
            'f_string': 'str-1',
            'two_1_1': self.ItemGeneralTwo(),
            'two_x_1': self.ItemGeneralTwo(),
            'two_1_x': two_1_x,
            'two_x_x': two_x_x,
        }
        item_one = self.ItemGeneralOne(**item_one_fields)
        
        for key in item_one_fields:
            self.assertIn(key, item_one)
        
        item_one.delete()
        
        for key in item_one_fields:
            self.assertNotIn(key, item_one)
        
        # --- bulk item container ---
        bulk_one = self.ItemGeneralOne.Bulk()
        bulk_one.gen()
        bulk_one.gen()
        
        self.assertEqual(len(bulk_one), 2)
        
        bulk_one.delete()
        
        self.assertEqual(len(bulk_one), 0)
        
        # --- bulk defaults must be cleared ---
        item_one_1 = self.ItemGeneralOne()
        bulk_one_1 = self.ItemGeneralOne.Bulk()
        
        bulk_two = self.ItemGeneralTwo.Bulk(
            one_x_1=item_one_1, one_x_x =bulk_one_1,
            f_integer=10, one_x_1__f_integer=20
        )
        
        self.assertIn('one_x_1', bulk_two)
        self.assertIn('one_x_x', bulk_two)
        self.assertIn('f_integer', bulk_two)
        self.assertIn('one_x_1__f_integer', bulk_two)
        
        bulk_two.delete()
        
        self.assertNotIn('one_x_1', bulk_two)
        self.assertNotIn('one_x_x', bulk_two)
        self.assertNotIn('f_integer', bulk_two)
        self.assertNotIn('one_x_1__f_integer', bulk_two)
        
        # --- deleted items must not be referenced ---
        # x-to-1
        item_one_1 = self.ItemGeneralOne()
        bulk_one_1 = self.ItemGeneralOne.Bulk()
        
        bulk_two = self.ItemGeneralTwo.Bulk(
            one_x_1=item_one_1, one_x_x =bulk_one_1,
            f_integer=10, one_x_1__f_integer=20
        )
        
        item_one_1.delete()
        
        self.assertNotIn('one_x_1', bulk_two)
        self.assertIs(bulk_two['one_x_x'], bulk_one_1)
        self.assertEqual(bulk_two['f_integer'], 10)
        self.assertEqual(bulk_two['one_x_1__f_integer'], 20)
        
        # x-to-x
        item_one_1 = self.ItemGeneralOne()
        bulk_one_1 = self.ItemGeneralOne.Bulk()
        
        bulk_two = self.ItemGeneralTwo.Bulk(
            one_x_1=item_one_1, one_x_x =bulk_one_1,
            f_integer=10, one_x_1__f_integer=20
        )
        
        bulk_one_1.delete()
        
        self.assertNotIn('one_x_x', bulk_two)
        self.assertIs(bulk_two['one_x_1'], item_one_1)
        self.assertEqual(bulk_two['f_integer'], 10)
        self.assertEqual(bulk_two['one_x_1__f_integer'], 20)
        
        


from save_to_db.adapters.utils.adapter_manager import get_adapter_cls

from save_to_db.utils.test_base import TestBase



class TestConstraints(TestBase):
    """ Tests that an adapter correctly returns not null and unique fields. """
    
    ModelConstraintsOne = None
    ModelConstraintsTwo = None
    ModelConstraintsThree = None
    ModelConstraintsFour = None
    ModelConstraintsFive = None
    ModelConstraintsSix = None
    ModelConstraintsSelf = None

        
    def test_iter_required_fields(self):
        adapter_cls = get_adapter_cls(self.ModelConstraintsOne)
        composites = adapter_cls.COMPOSITE_KEYS_SUPPORTED
        
        expected_not_nulls = [
            [self.ModelConstraintsOne, {'f_string', 'f_text'}],
            [self.ModelConstraintsTwo, {'four_x_1'}],
            [self.ModelConstraintsThree, {'one_x_1'}],
            [self.ModelConstraintsFour,
                {'primary_one', 'primary_two'} if composites else set()],
            [self.ModelConstraintsFive, set()],
            [self.ModelConstraintsSix, {'five_1_1'}],
            [self.ModelConstraintsSelf, {'code',
                                         'parent_x_1', 'second_parent_1_1'}],
        ]
        
        for model_cls, expected_not_null_fields in expected_not_nulls:
            not_null_fields = set(adapter_cls.iter_required_fields(model_cls))
            self.assertEqual(not_null_fields, expected_not_null_fields,
                             model_cls)
    
    
    def __to_sorted_lists(self, iterable_of_iterables):
        result = []
        for iterable in iterable_of_iterables:
            element = list(iterable)
            element.sort()
            result.append(element)
        result.sort()
        return result
    
    
    def test_iter_unique_field_combinations(self):
        adapter_cls = get_adapter_cls(self.ModelConstraintsOne)
        composites = adapter_cls.COMPOSITE_KEYS_SUPPORTED
        
        expected_uniques = [
            [self.ModelConstraintsOne, [{'id'}, {'f_integer'}, {'f_string'},
                                        ['five_1_1']]],
            [self.ModelConstraintsTwo, [{'id'}]],
            [self.ModelConstraintsThree, [{'id'}]],
            [self.ModelConstraintsFour,
                [{'primary_one', 'primary_two'} if composites else {'id'},
                 {'f_integer', 'f_string'}, {'five_1_1'}]],
            [self.ModelConstraintsFive, [{'id'}, {'one_1_1'}, {'four_1_1'},
                                         {'six_1_1'}]],
            [self.ModelConstraintsSix, [{'id'}, {'f_integer', 'five_1_1'},
                                        {'five_1_1'}]],
            [self.ModelConstraintsSelf, [{'code'},
                                         {'first_parent_1_1'},
                                         {'first_child_1_1'},
                                         {'second_parent_1_1'},
                                         {'second_child_1_1'}]],
        ]
        for model_cls, expected in expected_uniques:
            actual = self.__to_sorted_lists(
                        list(adapter_cls.iter_unique_field_combinations(
                            model_cls))
                    )
            expected = self.__to_sorted_lists(expected)
            
            self.assertEqual(actual, expected, model_cls)
       
       
            